# Copyright (C) 2019 Battelle Memorial Institute
# file: case_merge.py
"""Combines GridLAB-D and agent files to run a multi-feeder TESP simulation

Public Functions:
    :merge_glm: combines GridLAB-D input files
    :merge_glm_dict: combines GridLAB-D metadata files
    :merge_agent_dict: combines the substation agent configuration files
    :merge_substation_yaml: combines the substation agent FNCS publish/subscribe files
    :merge_fncs_config: combines GridLAB-D FNCS publish/subscribe files
    :merge_gld_msg: combines GridLAB-D HELICS publish/subscribe configurations
    :merge_substation_msg: combines the substation agent HELICS publish/subscribe configurations
"""
import sys
import json
import tesp_support.helpers as helpers

if sys.platform == 'win32':
  pycall = 'python'
else:
  pycall = 'python3'

def merge_glm(target, sources, xfmva):
  """Combines GridLAB-D input files into target/target.glm. The source files must already exist.

  Args:
      target (str): the directory and root case name
      sources (list): list of feeder names in the target directory to merge
  """
  print ('combining', sources, 'glm files into', target)
  workdir = './' + target + '/'
  op = open (workdir + target + '.glm', 'w')
  inFirstFile = True
  firstHeadNode = ''
  finishedFirstSubstation = False
  for fdr in sources:
    with open (workdir + fdr + '.glm') as ip:
      inPreamble = True
      inSubstation = False
      inHELICS = False
      thisHeadNode = ''
      for line in ip:
        canWrite = True
        if 'filename Voltage_Dump' in line:
            line = '  filename Voltage_Dump_' + target + '.csv;'
        if 'filename Current_Dump' in line:
            line = '  filename Current_Dump_' + target + '.csv;'
        if '#ifdef USE_FNCS' in line:
            inSubstation = True
        if 'object helics_msg' in line:
            inHELICS = True
        if inSubstation == True: 
          if '  configure' in line:
            if inHELICS == False:
              line = '  configure ' + target + '_FNCS_Config.txt;'
          elif '  power_rating' in line:
            line = '  power_rating {:.2f};'.format (xfmva * 1e3)
          elif '  base_power' in line:
            line = '  base_power {:.2f};'.format (xfmva * 1e6)
          elif '  to ' in line:
            toks = line.split()
            thisHeadNode = toks[1][:-1]
            if len(firstHeadNode) < 1:
              firstHeadNode = thisHeadNode
        if inHELICS == True:
          if 'configure' in line:
            line = '  configure ' + target + '_HELICS_gld_msg.json;'
            inHELICS = False
        if (inSubstation == True) and ('object node' in line):
          inSubstation = False
          if finishedFirstSubstation == True:
            print ('object switch {', file=op)
            print ('  name tie_' + fdr + ';', file=op)
            print ('  phases ABCN;', file=op)
            print ('  from ' + firstHeadNode + ';', file=op)
            print ('  to ' + thisHeadNode + ';', file=op)
            print ('  status CLOSED;', file=op)
            print ('}', file=op)
          finishedFirstSubstation = True
        if (inSubstation == True) and (finishedFirstSubstation == True):
          canWrite = False
        if (inPreamble == True) and (inFirstFile == False):
          canWrite = False
        if canWrite == True:
          print (line.rstrip(), file=op)
        if '#endif' in line:
          if '&&&' in line:
            if 'end of common section for combining TESP cases' in line:
              inPreamble = False
    inFirstFile = False 
  op.close()

def key_present (val, ary):
  tok = val['key']
  for msg in ary:
    if tok == msg['key']:
      return True
  return False

def merge_gld_msg (target, sources):
  print ('combining', sources, 'HELICS GridLAB-D json files into', target)
  workdir = './' + target + '/'
  dict = {"name":"gld1", "period":1, "subscriptions":[], "publications":[]}
  subs = []
  pubs = []
  for fdr in sources:
    lp = open (workdir + fdr + '_HELICS_gld_msg.json').read()
    cfg = json.loads(lp)
    dict["name"] = cfg["name"]
    dict["period"] = cfg["period"]
    for pub in cfg["publications"]:
      if not key_present (pub, pubs):
        pubs.append (pub)
    for sub in cfg["subscriptions"]:
      if not key_present (sub, subs):
        subs.append (sub)
  dict["publications"] = pubs
  dict["subscriptions"] = subs
  dp = open (workdir + target + '_HELICS_gld_msg.json', 'w')
  json.dump (dict, dp, ensure_ascii=False, indent=2)
  dp.close()

def merge_substation_msg (target, sources):
  print ('combining', sources, 'HELICS Substation json files into', target)
  workdir = './' + target + '/'
  dict = {"name":"gld1", "period":1, "subscriptions":[], "publications":[]}
  subs = []
  pubs = []
  for fdr in sources:
    lp = open (workdir + fdr + '_HELICS_substation.json').read()
    cfg = json.loads(lp)
    dict["name"] = cfg["name"]
    dict["period"] = cfg["period"]
    for pub in cfg["publications"]:
      if not key_present (pub, pubs):
        pubs.append (pub)
    for sub in cfg["subscriptions"]:
      if not key_present (sub, subs):
        subs.append (sub)
  dict["publications"] = pubs
  dict["subscriptions"] = subs
  dp = open (workdir + target + '_HELICS_substation.json', 'w')
  json.dump (dict, dp, ensure_ascii=False, indent=2)
  dp.close()

def merge_glm_dict(target, sources, xfmva):
  """Combines GridLAB-D metadata files into target/target.json. The source files must already exist.

  Each constituent feeder has a new ID constructed from the NamePrefix + original base_feeder,
  then every child object on that feeder will have its feeder_id, originally network_node, 
  changed to match the new one.

  Args:
      target (str): the directory and root case name
      sources (list): list of feeder names in the target directory to merge
  """
  print ('combining', sources, 'GridLAB-D json files into', target)
  dict = {'bulkpower_bus' : 'TBD', \
          'FedName' : 'gld1', \
          'transformer_MVA' : xfmva, \
          'feeders' : {}, \
          'billingmeters' : {}, \
          'houses' : {}, \
          'inverters' : {}, \
          'capacitors' : {}, \
          'regulators' : {}}
  workdir = './' + target + '/'
  for fdr in sources:
    cp = open (fdr + '.json').read()
    comb_cfg = json.loads(cp)
    name_prefix = comb_cfg['BackboneFiles']['NamePrefix']

    lp = open (workdir + fdr + '_glm_dict.json').read()
    cfg = json.loads(lp)
    fdr_id = helpers.gld_strict_name (name_prefix + cfg['feeders']['network_node']['base_feeder'])
    print ('created new feeder id', fdr_id, 'for', fdr)
    dict['feeders'][fdr_id] = {'house_count':cfg['feeders']['network_node']['house_count'], \
                               'inverter_count':cfg['feeders']['network_node']['inverter_count'], \
                               'base_feeder':cfg['feeders']['network_node']['base_feeder']}
    for key in ['billingmeters', 'houses', 'inverters', 'capacitors', 'regulators']:
      for obj in cfg[key]:
        if 'feeder_id' in cfg[key][obj]:
          cfg[key][obj]['feeder_id'] = fdr_id
      dict[key].update(cfg[key])
  op = open (workdir + target + '_glm_dict.json', 'w')
  print (json.dumps(dict), file=op)
  op.close()

def merge_agent_dict(target, sources, xfmva):
  """Combines the substation agent configuration files into target/target.json. The source files must already exist.

  Args:
      target (str): the directory and root case name
      sources (list): list of feeder names in the target directory to merge
  """
  print ('combining', sources, 'agent json files into', target)
  dict = {'markets':{},'controllers':{},'dt':0.0,'GridLABD':''}
  workdir = './' + target + '/'
  for fdr in sources:
    lp = open (workdir + fdr + '_agent_dict.json').read()
    cfg = json.loads(lp)
    if dict['dt'] <= 0.0:
      dict['dt'] = cfg['dt']
    if len(dict['GridLABD']) < 1:
      dict['GridLABD'] = cfg['GridLABD']
    for key in ['markets', 'controllers']:
      dict[key].update(cfg[key])
  for mkt in dict['markets']:
    dict['markets'][mkt]['max_capacity_reference_bid_quantity'] = xfmva * 1000.0 * (5.0 / 3.0)
  op = open (workdir + target + '_agent_dict.json', 'w')
  print (json.dumps(dict), file=op)
  op.close()

def merge_substation_yaml(target, sources):
  """Combines GridLAB-D input files into target/target.yaml. The source files must already exist.

  Args:
      target (str): the directory and root case name
      sources (list): list of feeder names in the target directory to merge
  """
  print ('combining', sources, 'yaml files into', target)
  workdir = './' + target + '/'
  op = open (workdir + target + '_substation.yaml', 'w')
  inFirstFile = True 
  for fdr in sources:
    with open (workdir + fdr + '_substation.yaml') as ip:
      numListFalse = 0
      for line in ip:
        if numListFalse >= 2 or inFirstFile == True:
          print (line.rstrip(), file=op)
        if 'list: false' in line:
          numListFalse += 1
    inFirstFile = False 
  op.close()

def merge_fncs_config(target, sources):
  """Combines GridLAB-D input files into target/target.txt. The source feeders must already exist.

  Args:
      target (str): the directory and root case name
      sources (list): list of feeder names in the target directory to merge
  """
  print ('combining', sources, 'txt files into', target)
  workdir = './' + target + '/'
  op = open (workdir + target + '_FNCS_Config.txt', 'w')
  inFirstFile = True 
  for fdr in sources:
    with open (workdir + fdr + '_FNCS_Config.txt') as ip:
      numLocalWeather = 0
      for line in ip:
        if numLocalWeather >= 6 or inFirstFile == True:
          print (line.rstrip(), file=op)
        if 'localWeather' in line:
          numLocalWeather += 1
    inFirstFile = False 
  op.close()




# Copyright (C) 2018-2019 Battelle Memorial Institute
# file: TMY2EPW.py
"""Functions to convert Typical Meteorological Year data for EnergyPlus

The input must be in TMY2 format. If only TMY3 files are available,
see the TMY3toTMY2_ansi.c program distributed with TESP under the
support/weather/TMY2EPW/source_code directory.

Public Functions:
    :convert_tmy2_to_epw: Converts TMY2 file to EPW.  

"""

import sys
import numpy as np

def removeZero(x):
    """ helper function to strip leading '0' from a string
    """
    if x[0]=='0':
       x=x[1:]
    return x

abbreviations = [
    ['AL','Alabama','USA'],
    ['AK','Alaska','USA'],
    ['AS','American Samoa','ASM'],
    ['AZ','Arizona','USA'],
    ['AR','Arkansas','USA'],
    ['CA','California','USA'],
    ['CO','Colorado','USA'],
    ['CT','Connecticut','USA'],
    ['DE','Delaware','USA'],
    ['DC','District of Columbia','USA'],
    ['FM','Fed. States of Micronesia','FSM'],
    ['FL','Florida','USA'],
    ['GA','Georgia','USA'],
    ['GU','Guam','GUM'],
    ['HI','Hawaii','USA'],
    ['ID','Idaho','USA'],
    ['IL','Illinois','USA'],
    ['IN','Indiana','USA'],
    ['IA','Iowa','USA'],
    ['KS','Kansas','USA'],
    ['KY','Kentucky','USA'],
    ['LA','Louisiana','USA'],
    ['ME','Maine','USA'],
    ['MH','Marshall Islands','MHL'],
    ['MD','Maryland','USA'],
    ['MA','Massachusetts','USA'],
    ['MI','Michigan','USA'],
    ['MN','Minnesota','USA'],
    ['MS','Mississippi','USA'],
    ['MO','Missouri','USA'],
    ['MT','Montana','USA'],
    ['NE','Nebraska','USA'],
    ['NV','Nevada','USA'],
    ['NH','New Hampshire','USA'],
    ['NJ','New Jersey','USA'],
    ['NM','New Mexico','USA'],
    ['NY','New York','USA'],
    ['NC','North Carolina','USA'],
    ['ND','North Dakota','USA'],
    ['MP','Northern Mariana Is.','MNP'],
    ['OH','Ohio','USA'],
    ['OK','Oklahoma','USA'],
    ['OR','Oregon','USA'],
    ['PW','Palau','PLW'],
    ['PA','Pennsylvania','USA'],
    ['PR','Puerto Rico','PRI'],
    ['RI','Rhode Island','USA'],
    ['SC','South Carolina','USA'],
    ['SD','South Dakota','USA'],
    ['TN','Tennessee','USA'],
    ['TX','Texas','USA'],
    ['UT','Utah','USA'],
    ['VT','Vermont','USA'],
    ['VA','Virginia','USA'],
    ['VI','Virgin Islands','VIR'],
    ['WA','Washington','USA'],
    ['WV','West Virginia','USA'],
    ['WI','Wisconsin','USA'],
    ['WY','Wyoming','USA'],
    ['AB','Alberta','CAN'],
    ['BC','British Columbia','CAN'],
    ['MB','Manitoba','CAN'],
    ['NB','New Brunswick','CAN'],
    ['NF','Newfoundland','CAN'],
    ['NT','Northwest Territories','CAN'],
    ['NS','Nova Scotia','CAN'],
    ['NU','Nunavut','CAN'],
    ['ON','Ontario','CAN'],
    ['PE','Prince Edward Island','CAN'],
    ['QC','Quebec','CAN'],
    ['PQ','Quebec','CAN'],
    ['SK','Saskatchewan','CAN'],
    ['YT','Yukon','CAN'],
    ['YT','Yukon Territory','CAN']]

def convert_tmy2_to_epw (fileroot):
  """ Converts TMY2 to EPW

  Reads fileroot.tmy2 and writes fileroot.epw

  Args:
    fileroot (str): The input path and base file name, without extension
  """
  f=open(fileroot + '.tmy2','r')
  lines=f.readlines()
  f.close()

  newline=[]

  WBAN=lines[0][0:6]  # this is a 6-digit AWSMSC read from TMY2, not a 5-digit WBAN read from TMY2
#  print ('WBAN/MSC',WBAN,'!')

  city=lines[0][7:29]
#  print ('city',city,'!')

  state=lines[0][30:32]
#  print ('state',state,'!')
  for row in abbreviations:
    if row[0].lower().find(state.lower())!=-1:
      country=row[2]
#  print ('country',country,'!')
  tz=lines[0][33:36]
#  print ('tz',tz,'!')
  dir=lines[0][37]
  if dir.lower().find('n')!=-1:
     sim=''
  else:
     sim='-'
  deg=float(lines[0][39:41])
  min=round(float(lines[0][39:41])/60,1)
  lat=sim+str(deg+min)
#  print ('lat',lat,'!')
  dir=lines[0][45]
  if dir.lower().find('e')!=-1:
     sim=''
  else:
     sim='-'
  deg=float(lines[0][47:50])
  min=round(float(lines[0][51:53])/60,1)
  lot=sim+str(deg+min)
#  print ('lot',lot,'!')
  ele=lines[0][55:59].replace(' ','')
#  print ('ele',ele,'!')
  temp='LOCATION,'+str(city)+','+str(state)+','+str(country)+',TMY2,'+str(WBAN)+','+str(lat)+','+str(lot)+','+str(tz)+','+str(ele)
  newline.append(temp)

  newline.append('DESIGN CONDITIONS,0')
  newline.append('TYPICAL/EXTREME PERIODS,0')
  newline.append('GROUND TEMPERATURES,0')
  newline.append('HOLIDAYS/DAYLIGHT SAVINGS,No,0,0,0')
  newline.append('COMMENTS 1,TMY3-'+str(WBAN)+' -- WMO#')
  newline.append('COMMENTS 2,')
  temp=('DATA PERIODS,1,1,Data,Sunday,')
  month=removeZero(lines[1][3:5])
  day=removeZero(lines[1][5:7])
  temp=temp+str(month)+'/'+str(day)
  month=removeZero(lines[-1][3:5])
  day=removeZero(lines[-1][5:7])
  temp=temp+','+ str(month)+'/'+str(day)
  newline.append(temp)

  datasource=[71,77,82,88,9,21,27,33,39,45,51,57,93,98,61,65,104,111,126,131,136,140]

  for i in range(1,len(lines)-1):
      temp=''
      year=lines[i][1:3]
      if int(year) < 60:
        year='20'+year
      else:
        year='19'+year
      temp=temp+year
      month=removeZero(lines[i][3:5])
      temp=temp+','+month
      day=removeZero(lines[i][5:7])
      temp=temp+','+day
      hour=removeZero(lines[i][7:9])
      temp=temp+','+hour
      minute='0'
      temp=temp+','+minute
      data=''
      for j in range(len(datasource)):
        if datasource[j]!=9:
          data=data+lines[i][datasource[j]:datasource[j]+2]
        else:
          data=data+'?0'
#     if i==1:
#       print ('data',data,'!')
      temp=temp+','+data
      drybulb=float(lines[i][67:71])/10
      temp=temp+','+str(drybulb)
      dewpoi=float(lines[i][73:77])/10
      temp=temp+','+str(dewpoi)
      RelHum =float(lines[i][79:82])
      temp=temp+','+str(RelHum)
      AtmPre =float(lines[i][84:88])*100
      temp=temp+','+str(AtmPre)
      ExtHorRad=lines[i][9:13]
      temp=temp+','+str(ExtHorRad)
      ExtDirNorRad=lines[i][13:17]
      temp=temp+','+str(ExtDirNorRad)
      OpaSkyCov=float(lines[i][63:65])
  #    temp=temp+','+str(OpaSkyCov)
      skyemi=(0.787+0.764*np.log((dewpoi+273.0)/273.0))*(1+0.0224*OpaSkyCov+0.0035*OpaSkyCov*OpaSkyCov+0.00028*OpaSkyCov*OpaSkyCov*OpaSkyCov)
      HorzIRSky=skyemi* 5.6697/100000000*((drybulb+273)*(drybulb+273)*(drybulb+273)*(drybulb+273))
#     if i==1:
#       print ('HorzIRSky',HorzIRSky,'!')
      temp=temp+',{:.3f}'.format (HorzIRSky)
      GloHorzRad=float(lines[i][17:21])
      temp=temp+','+str(GloHorzRad)
      DirNormRad=float(lines[i][23:27])
      temp=temp+','+str(DirNormRad)
      DifHorzRad=float(lines[i][29:33])
      temp=temp+','+str(DifHorzRad)
      GloHorzIllum=float(lines[i][35:39])
      temp=temp+','+str(GloHorzIllum)
      DirNormIllum=float(lines[i][41:45])
      temp=temp+','+str(DirNormIllum)
      DifHorzIllum=float(lines[i][47:51])
      temp=temp+','+str(DifHorzIllum)
      ZenLum=float(lines[i][53:57])
      temp=temp+','+str(ZenLum)
      WindDir=float(lines[i][90:93])
      temp=temp+','+str(WindDir)
      WindSpd=float(lines[i][95:98])/10
      temp=temp+','+str(WindSpd)
      TotSkyCvr=float(lines[i][59:61])
      temp=temp+','+str(TotSkyCvr)
      OpaqSkyCvr=float(lines[i][63:65])
      temp=temp+','+str(OpaqSkyCvr)
      Visibility=float(lines[i][100:104])/10
      temp=temp+','+str(Visibility)
      Ceiling_Hgt=float(lines[i][106:111])
      temp=temp+','+str(Ceiling_Hgt)
      PresWeath=lines[i][113:123]
      if PresWeath.find('999999999')!=-1:
        PresWeathObs='9'
      else:
        PresWeathObs='0'
#     if i==1:
#       print ('PresWeath',PresWeath,'!')
#       print ('PresWeathObs',PresWeathObs,'!')
      temp=temp+','+str(PresWeathObs)
      temp=temp+','+str(PresWeath)
      Precip_Wtr=float(lines[i][123:126])
      temp=temp+','+str(Precip_Wtr)
      AerOptDep=float(lines[i][128:131])
      temp=temp+','+str(AerOptDep)
      SnoDep=float(lines[i][133:136])
      temp=temp+','+str(SnoDep)
      DayLasSno=float(lines[i][138:140])
      temp=temp+','+str(DayLasSno)
      Albedo='0'
      temp=temp+','+str(Albedo)
      Rain='0'
      temp=temp+','+str(Rain)
      RaiQua='0'
      temp=temp+','+str(RaiQua)
      newline.append(temp)

  f=open(fileroot + '.epw','w')

  for i in range(len(newline)):
    f.writelines(newline[i]+'\n')
  f.close()


# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: process_inv.py
"""Functions to plot inverter and volt-var data from GridLAB-D, for NIST TE Challenge 2

Public Functions:
    :process_inv: Reads the data and metadata, then makes the plots.  

"""
import json;
import sys;
import numpy as np;
import os.path;
try:
  import matplotlib as mpl;
  import matplotlib.pyplot as plt;
except:
  pass

def process_inv(nameroot, dictname = ''):
  """ Plots inverter and volt-var data for the NIST TE Challenge 2 / IEEE 8500 examples

  This function reads *substation_nameroot_metrics.json*,  
  *billing_meter_nameroot_metrics.json*,
  *capacitor_nameroot_metrics.json*,
  *regulator_nameroot_metrics.json*,
  *house_nameroot_metrics.json* and
  *inverter_nameroot_metrics.json* for the data;
  it reads *nameroot_glm_dict.json* for the metadata.
  If possible, it reads *precool_nameroot_metrics.json* for temperature deviation.  
  These must all exist in the current working directory.  
  One graph is generated with 10 subplots:
  
  1. Average P and Q over all inverters
  2. Min, Max and Average line-neutral voltage over all billing meters
  3. Average air temperature over all houses
  4. Average temperature deviations from the setpoint over all houses
  5. Total of ANSI C84 A and B range violation counts, summing over all billing meters
  6. Total of ANSI C84 A and B range violation durations, summing over all billing meters
  7. Substation total power, losses, house power, house HVAC power and house waterheater power
  8. The accumulated bill, summed over all billing meters
  9. The accumulated capacitor switching counts for each of 4 capacitor banks, if found, as in the IEEE 8500 case
  10. The accumulated regulator counts for each of 4 voltage regulators, if found, as in the IEEE 8500 case

  Args:
    nameroot (str): name of the TESP case, not necessarily the same as the GLM case, without the extension
    dictname (str): metafile name (with json extension) for a different GLM dictionary, if it's not *nameroot_glm_dict.json*. Defaults to empty.
  """
  # first, read and print a dictionary of all the monitored GridLAB-D objects
  if len (dictname) > 0:
      lp = open (dictname).read()
  else:
      lp = open (nameroot + "_glm_dict.json").read()
  dict = json.loads(lp)
  sub_keys = list(dict['feeders'].keys())
  sub_keys.sort()
  inv_keys = list(dict['inverters'].keys())
  inv_keys.sort()
  hse_keys = list(dict['houses'].keys())
  hse_keys.sort()
  mtr_keys = list(dict['billingmeters'].keys())
  mtr_keys.sort()
  cap_keys = list(dict['capacitors'].keys())
  cap_keys.sort()
  reg_keys = list(dict['regulators'].keys())
  reg_keys.sort()
  xfMVA = dict['transformer_MVA']
  bulkBus = dict['bulkpower_bus']
  #print ("\n\nFile", nameroot, "has substation", sub_keys[0], "at bulk system bus", bulkBus, "with", xfMVA, "MVA transformer")
  #print("\nFeeder Dictionary:")
  #for key in sub_keys:
  # row = dict['feeders'][key]
  # print (key, "has", row['house_count'], "houses and", row['inverter_count'], "inverters")
  #print("\nBilling Meter Dictionary:")
  #for key in mtr_keys:
  # row = dict['billingmeters'][key]
  # print (key, "on phase", row['phases'], "of", row['feeder_id'], "with", row['children'])
  #print("\nHouse Dictionary:")
  #for key in hse_keys:
  # row = dict['houses'][key]
  # print (key, "on", row['billingmeter_id'], "has", row['sqft'], "sqft", row['cooling'], "cooling", row['heating'], "heating", row['wh_gallons'], "gal WH")
  # # row['feeder_id'] is also available
  #print("\nInverter Dictionary:")
  #for key in inv_keys:
  # row = dict['inverters'][key]
  # print (key, "on", row['billingmeter_id'], "has", row['rated_W'], "W", row['resource'], "resource")
  # # row['feeder_id'] is also available

  # parse the substation metrics file first; there should just be one entity per time sample
  # each metrics file should have matching time points
  lp_s = open ("substation_" + nameroot + "_metrics.json").read()
  lst_s = json.loads(lp_s)
  print ("\nMetrics data starting", lst_s['StartTime'])

  # make a sorted list of the sample times in hours
  lst_s.pop('StartTime')
  meta_s = lst_s.pop('Metadata')
  times = list(map(int,list(lst_s.keys())))
  times.sort()
  print ("There are", len (times), "sample times at", times[1] - times[0], "second intervals")
  hrs = np.array(times, dtype=np.float)
  denom = 3600.0
  hrs /= denom

  time_key = str(times[0])

  # parse the substation metadata for 2 things of specific interest
  #print ("\nSubstation Metadata for", len(lst_s[time_key]), "objects")
  for key, val in meta_s.items():
  # print (key, val['index'], val['units'])
    if key == 'real_power_avg':
      SUB_POWER_IDX = val['index']
      SUB_POWER_UNITS = val['units']
    elif key == 'real_power_losses_avg':
      SUB_LOSSES_IDX = val['index']
      SUB_LOSSES_UNITS = val['units']

  # create a NumPy array of all metrics for the substation
  data_s = np.empty(shape=(len(sub_keys), len(times), len(lst_s[time_key][sub_keys[0]])), dtype=np.float)
  #print ("\nConstructed", data_s.shape, "NumPy array for Substations")
  j = 0
  for key in sub_keys:
    i = 0
    for t in times:
      ary = lst_s[str(t)][sub_keys[j]]
      data_s[j, i,:] = ary
      i = i + 1
    j = j + 1

  # read the other JSON files; their times (hrs) should be the same
  lp_h = open ("house_" + nameroot + "_metrics.json").read()
  lst_h = json.loads(lp_h)
  lp_m = open ("billing_meter_" + nameroot + "_metrics.json").read()
  lst_m = json.loads(lp_m)
  lp_i = open ("inverter_" + nameroot + "_metrics.json").read()
  lst_i = json.loads(lp_i)
  lp_c = open ("capacitor_" + nameroot + "_metrics.json").read()
  lst_c = json.loads(lp_c)
  lp_r = open ("regulator_" + nameroot + "_metrics.json").read()
  lst_r = json.loads(lp_r)

  # houses
  lst_h.pop('StartTime')
  meta_h = lst_h.pop('Metadata')
  #print("\nHouse Metadata for", len(lst_h[time_key]), "objects")
  for key, val in meta_h.items():
  # print (key, val['index'], val['units'])
    if key == 'air_temperature_max':
      HSE_AIR_MAX_IDX = val['index']
      HSE_AIR_MAX_UNITS = val['units']
    elif key == 'air_temperature_min':
      HSE_AIR_MIN_IDX = val['index']
      HSE_AIR_MIN_UNITS = val['units']
    elif key == 'air_temperature_avg':
      HSE_AIR_AVG_IDX = val['index']
      HSE_AIR_AVG_UNITS = val['units']
    elif key == 'air_temperature_median':
      HSE_AIR_MED_IDX = val['index']
      HSE_AIR_MED_UNITS = val['units']
    elif key == 'total_load_avg':
      HSE_TOTAL_AVG_IDX = val['index']
      HSE_TOTAL_AVG_UNITS = val['units']
    elif key == 'hvac_load_avg':
      HSE_HVAC_AVG_IDX = val['index']
      HSE_HVAC_AVG_UNITS = val['units']
    elif key == 'waterheater_load_avg':
      HSE_WH_AVG_IDX = val['index']
      HSE_WH_AVG_UNITS = val['units']

  data_h = np.empty(shape=(len(hse_keys), len(times), len(lst_h[time_key][hse_keys[0]])), dtype=np.float)
  #print ("\nConstructed", data_h.shape, "NumPy array for Houses")
  j = 0
  for key in hse_keys:
    i = 0
    for t in times:
      ary = lst_h[str(t)][hse_keys[j]]
      data_h[j, i,:] = ary
      i = i + 1
    j = j + 1

  # Billing Meters 
  lst_m.pop('StartTime')
  meta_m = lst_m.pop('Metadata')
  #print("\nBilling Meter Metadata for", len(lst_m[time_key]), "objects")
  for key, val in meta_m.items():
  # print (key, val['index'], val['units'])
    if key == 'voltage_max':
      MTR_VOLT_MAX_IDX = val['index']
      MTR_VOLT_MAX_UNITS = val['units']
    elif key == 'voltage_min':
      MTR_VOLT_MIN_IDX = val['index']
      MTR_VOLT_MIN_UNITS = val['units']
    elif key == 'voltage_avg':
      MTR_VOLT_AVG_IDX = val['index']
      MTR_VOLT_AVG_UNITS = val['units']
    elif key == 'voltage12_max':
      MTR_VOLT12_MAX_IDX = val['index']
      MTR_VOLT12_MAX_UNITS = val['units']
    elif key == 'voltage12_min':
      MTR_VOLT12_MIN_IDX = val['index']
      MTR_VOLT12_MIN_UNITS = val['units']
    elif key == 'voltage_unbalance_max':
      MTR_VOLTUNB_MAX_IDX = val['index']
      MTR_VOLTUNB_MAX_UNITS = val['units']
    elif key == 'real_energy':
      ENERGY_IDX = val['index']
    elif key == 'bill':
      MTR_BILL_IDX = val['index']
      MTR_BILL_UNITS = val['units']
    elif key == 'above_RangeA_Count':
      MTR_AHI_COUNT_IDX = val['index']
    elif key == 'above_RangeB_Count':
      MTR_BHI_COUNT_IDX = val['index']
    elif key == 'below_RangeA_Count':
      MTR_ALO_COUNT_IDX = val['index']
    elif key == 'below_RangeB_Count':
      MTR_BLO_COUNT_IDX = val['index']
    elif key == 'below_10_percent_NormVol_Count':
      MTR_OUT_COUNT_IDX = val['index']
    elif key == 'above_RangeA_Duration':
      MTR_AHI_DURATION_IDX = val['index']
    elif key == 'above_RangeB_Duration':
      MTR_BHI_DURATION_IDX = val['index']
    elif key == 'below_RangeA_Duration':
      MTR_ALO_DURATION_IDX = val['index']
    elif key == 'below_RangeB_Duration':
      MTR_BLO_DURATION_IDX = val['index']
    elif key == 'below_10_percent_NormVol_Duration':
      MTR_OUT_DURATION_IDX = val['index']

  data_m = np.empty(shape=(len(mtr_keys), len(times), len(lst_m[time_key][mtr_keys[0]])), dtype=np.float)
  #print ("\nConstructed", data_m.shape, "NumPy array for Meters")
  j = 0
  for key in mtr_keys:
    i = 0
    for t in times:
      ary = lst_m[str(t)][mtr_keys[j]]
      data_m[j, i,:] = ary
      i = i + 1
    j = j + 1

  # Inverters 
  lst_i.pop('StartTime')
  meta_i = lst_i.pop('Metadata')
  #print("\nInverter Metadata for", len(lst_i[time_key]), "objects")
  for key, val in meta_i.items():
  # print (key, val['index'], val['units'])
    if key == 'real_power_avg':
      INV_P_AVG_IDX = val['index']
      INV_P_AVG_UNITS = val['units']
    elif key == 'reactive_power_avg':
      INV_Q_AVG_IDX = val['index']
      INV_Q_AVG_UNITS = val['units']

  data_i = np.empty(shape=(len(inv_keys), len(times), len(lst_i[time_key][inv_keys[0]])), dtype=np.float)
  print ("\nConstructed", data_i.shape, "NumPy array for Inverters")
  j = 0
  for key in inv_keys:
    i = 0
    for t in times:
      ary = lst_i[str(t)][inv_keys[j]]
      data_i[j, i,:] = ary
      i = i + 1
    j = j + 1

  # Precooling: won't necessarily have the same times?
  fname_p = 'precool_' + nameroot + '_metrics.json'
  have_precool = False
  if os.path.exists(fname_p):
    have_precool = True
    lp_p = open (fname_p).read()
    lst_p = json.loads(lp_p)
    lst_p.pop('StartTime')
    meta_p = lst_p.pop('Metadata')
    times_p = list(map(int,list(lst_p.keys())))
    times_p.sort()
    print ("There are", len (times_p), "agent sample times at", times_p[1] - times_p[0], "second intervals")
    hrs_p = np.array(times_p, dtype=np.float)
    denom = 3600.0
    hrs_p /= denom
    time_p_key = str(times_p[0])
    for key, val in meta_p.items():
      if key == 'temperature_deviation_avg':
        TEMPDEV_AVG_IDX = val['index']
        TEMPDEV_AVG_UNITS = val['units']
      elif key == 'temperature_deviation_min':
        TEMPDEV_MIN_IDX = val['index']
        TEMPDEV_MIN_UNITS = val['units']
      elif key == 'temperature_deviation_max':
        TEMPDEV_MAX_IDX = val['index']
        TEMPDEV_MAX_UNITS = val['units']
    data_p = np.empty(shape=(1, len(times_p), len(lst_p[time_p_key])), dtype=np.float)
    print ("\nConstructed", data_p.shape, "NumPy array for Agents")
    for i, t in enumerate (times_p):
      ary = lst_p[str(t)]
      data_p[0, i,:] = ary

  have_caps = False
  have_regs = False

  # Capacitors
  if len(cap_keys) > 0:
    have_caps = True
    lst_c.pop('StartTime')
    meta_c = lst_c.pop('Metadata')
#    print("\nCapacitor Metadata for", len(lst_c[time_key]), "objects")
    for key, val in meta_c.items():
      if key == 'operation_count':
        CAP_COUNT_IDX = val['index']
        CAP_COUNT_UNITS = val['units']
    data_c = np.empty(shape=(len(cap_keys), len(times), len(lst_c[time_key][cap_keys[0]])), dtype=np.float)
    print ("\nConstructed", data_c.shape, "NumPy array for Capacitors")
    j = 0
    for key in cap_keys:
      i = 0
      for t in times:
        ary = lst_c[str(t)][cap_keys[j]]
        data_c[j, i,:] = ary
        i = i + 1
      j = j + 1

  # Regulators
  if len(reg_keys) > 0:
    have_regs = True
    lst_r.pop('StartTime')
    meta_r = lst_r.pop('Metadata')
#    print("\nRegulator Metadata for", len(lst_r[time_key]), "objects")
    for key, val in meta_r.items():
      if key == 'operation_count':
        REG_COUNT_IDX = val['index']
        REG_COUNT_UNITS = val['units']
    data_r = np.empty(shape=(len(reg_keys), len(times), len(lst_r[time_key][reg_keys[0]])), dtype=np.float)
    print ("\nConstructed", data_r.shape, "NumPy array for Regulators")
    j = 0
    for key in reg_keys:
      i = 0
      for t in times:
        ary = lst_r[str(t)][reg_keys[j]]
        data_r[j, i,:] = ary
        i = i + 1
      j = j + 1

  ## assemble the total solar and battery inverter power
  j = 0
  solar_kw = np.zeros(len(times), dtype=np.float)
  battery_kw = np.zeros(len(times), dtype=np.float)
  for key in inv_keys:
    res = dict['inverters'][key]['resource']
    if res == 'solar':
      solar_kw += 0.001 * data_i[j,:,INV_P_AVG_IDX]
    elif res == 'battery':
      battery_kw += 0.001 * data_i[j,:,INV_P_AVG_IDX]
    j = j + 1

  hour1 = 4.0
  ihour1 = 0
  ihour1_p = 0
  for i in range(0, len(hrs)):
    if hrs[i] > hour1:
      ihour1 = i
      break

  if have_precool:
    for i in range(0, len(hrs_p)):
      if hrs_p[i] > hour1:
        ihour1_p = i
        break

  # display some averages
  print ("Maximum feeder power =", '{:.2f}'.format(0.001*data_s[0,:,SUB_POWER_IDX].max()), 'kW')
  print ("Average feeder power =", '{:.2f}'.format(0.001*data_s[0,:,SUB_POWER_IDX].mean()), 'kW')
  print ("Average feeder losses =", '{:.2f}'.format(0.001*data_s[0,:,SUB_LOSSES_IDX].mean()), 'kW')
  print ('Average all house temperatures Noon-8 pm day 1:', '{:.2f}'.format(data_h[:,144:240,HSE_AIR_AVG_IDX].mean()))
  #print ('Average all house temperatures Noon-8 pm day 2:', '{:.2f}'.format(data_h[:,432:528,HSE_AIR_AVG_IDX].mean()))
  print ("Average inverter P =", '{:.2f}'.format(data_i[:,:,INV_P_AVG_IDX].mean()), INV_P_AVG_UNITS)
  print ("Average inverter Q =", '{:.2f}'.format(data_i[:,:,INV_Q_AVG_IDX].mean()), INV_Q_AVG_UNITS)
  print ("A Range Hi Duration =", '{:.2f}'.format(data_m[:,:,MTR_AHI_DURATION_IDX].sum() / 3600.0), 
         "count =", '{:.2f}'.format(data_m[:,:,MTR_AHI_COUNT_IDX].sum()))
  print ("A Range Lo Duration =", '{:.2f}'.format(data_m[:,:,MTR_ALO_DURATION_IDX].sum() / 3600.0), 
         "count =", '{:.2f}'.format(data_m[:,:,MTR_ALO_COUNT_IDX].sum()))
  print ("B Range Hi Duration =", '{:.2f}'.format(data_m[:,:,MTR_BHI_DURATION_IDX].sum() / 3600.0), 
         "count =", '{:.2f}'.format(data_m[:,:,MTR_BHI_COUNT_IDX].sum()))
  print ("B Range Lo Duration =", '{:.2f}'.format(data_m[:,:,MTR_BLO_DURATION_IDX].sum() / 3600.0), 
         "count =", '{:.2f}'.format(data_m[:,:,MTR_BLO_COUNT_IDX].sum()))
  print ("Zero-Volts Duration =", '{:.2f}'.format(data_m[:,:,MTR_OUT_DURATION_IDX].sum() / 3600.0), 
         "count =", '{:.2f}'.format(data_m[:,:,MTR_OUT_COUNT_IDX].sum()))
  if have_caps:
    print ("Total cap switchings =", '{:.2f}'.format(data_c[:,-1,CAP_COUNT_IDX].sum()))
  if have_regs:
    print ("Total tap changes =", '{:.2f}'.format(data_r[:,-1,REG_COUNT_IDX].sum()))

  final_bill = np.empty(shape=(len(times)), dtype=np.float)
  final_bill[0] = 0.0
  for i in range (1, len(hrs)):
    if hrs[i] > 15.0 and hrs[i] <= 19.0:
      price = 0.15
    else:
      price = 0.11
    kwh = 0.001 * data_m[:,i,ENERGY_IDX].sum()
#    print ('adding', kwh, 'at', price)
    final_bill[i] = final_bill[i-1] + price * kwh

  monthly_fee = 10.0  # the fixed charge from TEPCO used in NIST TE Challenge; not te30
  total_monthly = monthly_fee * len(mtr_keys)
  print ('Total Fixed Montly={:.2f}'.format (total_monthly))

  print ("Initial meter bill =", '{:.2f}'.format(data_m[:,-1,MTR_BILL_IDX].sum() - total_monthly))
  print ("Final meter bill =", '{:.2f}'.format(final_bill[-1]))
  if have_precool:
    print ("Average Temperature Deviation =", '{:.2f}'.format(data_p[:,:,TEMPDEV_AVG_IDX].mean()))

  print ('Summarizing from', hour1, 'hours to begin at indices', ihour1, ihour1_p)
  print ("Interval A Range Hi Duration =", '{:.2f}'.format(data_m[:,ihour1:-1,MTR_AHI_DURATION_IDX].sum() / 3600.0))
  print ("Interval A Range Lo Duration =", '{:.2f}'.format(data_m[:,ihour1:-1,MTR_ALO_DURATION_IDX].sum() / 3600.0))
  print ("Interval B Range Hi Duration =", '{:.2f}'.format(data_m[:,ihour1:-1,MTR_BHI_DURATION_IDX].sum() / 3600.0))
  print ("Interval B Range Lo Duration =", '{:.2f}'.format(data_m[:,ihour1:-1,MTR_BLO_DURATION_IDX].sum() / 3600.0)) 
  if have_precool:
    print ("Interval Average Temperature Deviation =", '{:.2f}'.format(data_p[:,ihour1_p:-1,TEMPDEV_AVG_IDX].mean()))
  if have_caps:
    print ("Interval Cap Switchings =", '{:.2f}'.format(data_c[:,-1,CAP_COUNT_IDX].sum() - data_c[:,ihour1,CAP_COUNT_IDX].sum()))
  if have_regs:
    print ("Interval Tap Changes =", '{:.2f}'.format(data_r[:,-1,REG_COUNT_IDX].sum() - data_r[:,ihour1,REG_COUNT_IDX].sum()))


  # create summary arrays
  total1 = (data_h[:,:,HSE_TOTAL_AVG_IDX]).squeeze()
  total2 = total1.sum(axis=0)
  hvac1 = (data_h[:,:,HSE_HVAC_AVG_IDX]).squeeze()
  hvac2 = hvac1.sum(axis=0)
  wh1 = (data_h[:,:,HSE_WH_AVG_IDX]).squeeze()
  wh2 = wh1.sum(axis=0)
  subkw = 0.001 * data_s[0,:,SUB_POWER_IDX]
  losskw = 0.001 * data_s[0,:,SUB_LOSSES_IDX]
  pavg1 = (data_i[:,:,INV_P_AVG_IDX]).squeeze()
  pavg2 = 0.001 * pavg1.mean(axis=0)
  qavg1 = (data_i[:,:,INV_Q_AVG_IDX]).squeeze()
  qavg2 = 0.001 * qavg1.mean(axis=0)
  tavg1 = (data_h[:,:,HSE_AIR_AVG_IDX]).squeeze()
  tavg2 = tavg1.mean(axis=0)

  vavg = np.zeros(shape=(len(times)), dtype=np.float)
  vmin = np.full(shape=(len(times)), fill_value=1000.0, dtype=np.float)
  vmax = np.zeros(shape=(len(times)), dtype=np.float)
  for i, mtr in enumerate(mtr_keys):
    vbase = dict['billingmeters'][mtr]['vln']
    vscale = 100.0 / vbase
    vavg += vscale * data_m[i,:,MTR_VOLT_AVG_IDX]
    vmin = np.minimum (vmin, vscale * data_m[i,:,MTR_VOLT_MIN_IDX])
    vmax = np.maximum (vmax, vscale * data_m[i,:,MTR_VOLT_MAX_IDX])
  vavg /= float (len(mtr_keys))

  # display a plot

  tmin = 0.0
  tmax = 24.0
  xticks = [0,4,8,12,16,20,24]

  SMALL_SIZE = 10
  MEDIUM_SIZE = 12
  BIGGER_SIZE = 14

  plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
  plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
  plt.rc('axes', labelsize=SMALL_SIZE)    # fontsize of the x and y labels
  plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
  plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
  plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
  plt.rc('figure', titlesize=SMALL_SIZE)  # fontsize of the figure title

  if have_caps or have_regs:
    fig, ax = plt.subplots(2, 5, sharex = 'col', figsize=(14,7))
  else:
    fig, ax = plt.subplots(2, 4, sharex = 'col', figsize=(14,6))

  ax[0,0].plot(hrs, pavg2, color="blue", label="P")
  ax[0,0].plot(hrs, qavg2, color="red", label="Q")
  ax[0,0].set_ylabel("kVA")
  ax[0,0].set_title ("Average Inverter Power", size=MEDIUM_SIZE)
  ax[0,0].legend(loc='best')
  ax[0,0].set_xlim(tmin,tmax)
  ax[0,0].set_xticks(xticks)

  #vabase = dict['inverters'][inv_keys[0]]['rated_W']
  #print ("Inverter base power =", vabase)
  #ax[0,1].plot(hrs, data_i[0,:,INV_P_AVG_IDX] / vabase, color="blue", label="Real")
  #ax[0,1].plot(hrs, data_i[0,:,INV_Q_AVG_IDX] / vabase, color="red", label="Reactive")
  #ax[0,1].set_ylabel("perunit")
  #ax[0,1].set_title ("Inverter Power at " + inv_keys[0])
  #ax[0,1].legend(loc='best')

  ax[1,0].plot(hrs, vmax, color="blue", label="Max")
  ax[1,0].plot(hrs, vmin, color="red", label="Min")
  ax[1,0].plot(hrs, vavg, color="green", label="Avg")
  ax[1,0].set_xlabel("Hours")
  ax[1,0].set_ylabel("%")
  ax[1,0].set_title ("All Meter Voltages", size=MEDIUM_SIZE)
  ax[1,0].legend(loc='best')
  ax[1,0].set_xlim(tmin,tmax)
  ax[1,0].set_xticks(xticks)

  ax[0,1].plot(hrs, tavg2, color="red", label="Avg")
  ax[0,1].set_ylabel('degF')
  ax[0,1].set_title ('Average House Temperatures', size=MEDIUM_SIZE)
  ax[0,1].set_xlim(tmin,tmax)
  ax[0,1].set_xticks(xticks)

  if have_precool:
    ax[1,1].plot(hrs_p, data_p[0,:,TEMPDEV_AVG_IDX], color="blue", label="Mean")
    #ax[1,1].plot(hrs_p, data_p[0,:,TEMPDEV_MIN_IDX], color="red", label="Min")
    #ax[1,1].plot(hrs_p, data_p[0,:,TEMPDEV_MAX_IDX], color="green", label="Max")
    ax[1,1].set_xlabel("Hours")
    ax[1,1].set_ylabel(TEMPDEV_AVG_UNITS)
    ax[1,1].set_title ("Average Temperature Deviations", size=MEDIUM_SIZE)
    #ax[1,1].legend(loc='best')
    ax[1,1].set_xlim(tmin,tmax)
    ax[1,1].set_xticks(xticks)
  else:
    ax[1,1].set_title ('No Thermostat Agents')


  ax[0,2].plot(hrs, (data_m[:,:,MTR_AHI_COUNT_IDX]).squeeze().sum(axis=0), color="blue", label="Range A Hi")
  ax[0,2].plot(hrs, (data_m[:,:,MTR_BHI_COUNT_IDX]).squeeze().sum(axis=0), color="cyan", label="Range B Hi")
  ax[0,2].plot(hrs, (data_m[:,:,MTR_ALO_COUNT_IDX]).squeeze().sum(axis=0), color="green", label="Range A Lo")
  ax[0,2].plot(hrs, (data_m[:,:,MTR_BLO_COUNT_IDX]).squeeze().sum(axis=0), color="magenta", label="Range B Lo")
  ax[0,2].plot(hrs, (data_m[:,:,MTR_OUT_COUNT_IDX]).squeeze().sum(axis=0), color="red", label="No Voltage")
  ax[0,2].set_ylabel("")
  ax[0,2].set_title ("All Voltage Violation Counts", size=MEDIUM_SIZE)
  ax[0,2].legend(loc='best')
  ax[0,2].set_xlim(tmin,tmax)
  ax[0,2].set_xticks(xticks)

  scalem = 1.0 / 3600.0
  ax[1,2].plot(hrs, scalem * (data_m[:,:,MTR_AHI_DURATION_IDX]).squeeze().sum(axis=0), color="blue", label="Range A Hi")
  ax[1,2].plot(hrs, scalem * (data_m[:,:,MTR_BHI_DURATION_IDX]).squeeze().sum(axis=0), color="cyan", label="Range B Hi")
  ax[1,2].plot(hrs, scalem * (data_m[:,:,MTR_ALO_DURATION_IDX]).squeeze().sum(axis=0), color="green", label="Range A Lo")
  ax[1,2].plot(hrs, scalem * (data_m[:,:,MTR_BLO_DURATION_IDX]).squeeze().sum(axis=0), color="magenta", label="Range B Lo")
  ax[1,2].plot(hrs, scalem * (data_m[:,:,MTR_OUT_DURATION_IDX]).squeeze().sum(axis=0), color="red", label="No Voltage")
  ax[1,2].set_xlabel("Hours")
  ax[1,2].set_ylabel("Hours")
  ax[1,2].set_title ("All Voltage Violation Durations", size=MEDIUM_SIZE)
  ax[1,2].legend(loc='best')
  ax[1,2].set_xlim(tmin,tmax)
  ax[1,2].set_xticks(xticks)

  ax[0,3].plot(hrs, subkw, color="blue", label="Substation")
  ax[0,3].plot(hrs, losskw, color="red", label="Losses")
  ax[0,3].plot(hrs, total2, color="green", label="Houses")
  ax[0,3].plot(hrs, hvac2, color="magenta", label="HVAC")
  ax[0,3].plot(hrs, wh2, color="orange", label="WH")
  ax[0,3].set_ylabel('kW')
  ax[0,3].set_title ("Average Real Power", size=MEDIUM_SIZE)
  ax[0,3].legend(loc='best')
  ax[0,3].set_xlim(tmin,tmax)
  ax[0,3].set_xticks(xticks)

  #ax[1,3].plot(hrs, data_m[0,:,MTR_BILL_IDX], color="blue")
  ax[1,3].plot(hrs, (data_m[:,:,MTR_BILL_IDX]).squeeze().sum(axis=0) - total_monthly, color='blue', label='Tariff')
  ax[1,3].plot(hrs, final_bill, color='red', label='Dynamic')
  ax[1,3].set_xlabel("Hours")
  ax[1,3].set_ylabel(MTR_BILL_UNITS)
  ax[1,3].set_title ("Meter Bills", size=MEDIUM_SIZE)
  ax[1,3].legend(loc='best')
  ax[1,3].set_xlim(tmin,tmax)
  ax[1,3].set_xticks(xticks)

  if have_caps:
    ax[0,4].plot(hrs, data_c[0,:,CAP_COUNT_IDX], color="blue", label=cap_keys[0])
    ax[0,4].plot(hrs, data_c[1,:,CAP_COUNT_IDX], color="red", label=cap_keys[1])
    ax[0,4].plot(hrs, data_c[2,:,CAP_COUNT_IDX], color="green", label=cap_keys[2])
    ax[0,4].plot(hrs, data_c[3,:,CAP_COUNT_IDX], color="magenta", label=cap_keys[3])
    ax[0,4].set_ylabel("")
    ax[0,4].set_title ("Cap Switchings", size=MEDIUM_SIZE)
    ax[0,4].legend(loc='best')
    ax[0,4].set_xlim(tmin,tmax)
    ax[0,4].set_xticks(xticks)

  if have_regs:
    ax[1,4].plot(hrs, data_r[0,:,REG_COUNT_IDX], color="blue", label=reg_keys[0])
    ax[1,4].plot(hrs, data_r[1,:,REG_COUNT_IDX], color="red", label=reg_keys[1])
    ax[1,4].plot(hrs, data_r[2,:,REG_COUNT_IDX], color="green", label=reg_keys[2])
    ax[1,4].plot(hrs, data_r[3,:,REG_COUNT_IDX], color="magenta", label=reg_keys[3])
    ax[1,4].set_xlabel("Hours")
    ax[1,4].set_ylabel("")
    ax[1,4].set_title ("Tap Changes", size=MEDIUM_SIZE)
    ax[1,4].legend(loc='best')
    ax[1,4].set_xlim(tmin,tmax)
    ax[1,4].set_xticks(xticks)

  if have_caps or have_regs:
    ax[1,4].set_xlabel("Hours")

  plt.tight_layout(pad=1.0, w_pad=1.0, h_pad=2.0)
  plt.show()





# Copyright (C) 2018-2020 Battelle Memorial Institute
# file: prep_precool.py
"""Writes the precooling agent and GridLAB-D metadata for NIST TE Challenge 2 example
 
Public Functions:
    :prep_precool: writes the JSON and YAML files 
"""
import sys
import json
import numpy as np

def prep_precool (nameroot, time_step=15):
    """Sets up agent configurations for the NIST TE Challenge 2 example

    Reads the GridLAB-D data from nameroot.glm; it should contain 
    houses with thermal_integrity_level attributes. Writes:

    - *nameroot_agent_dict.json*, contains configuration data for the precooler agents
    - *nameroot_precool.yaml*, contains FNCS subscriptions for the precooler agents
    - *nameroot_FNCS_Config.txt*, a GridLAB-D include file with FNCS publications and subscriptions

    Args:
        nameroot (str): the name of the GridLAB-D file, without extension
    """
    # we want the same psuedo-random thermostat schedules each time, for repeatability
    np.random.seed (0)

    # write yaml for precool.py to subscribe meter voltages and house setpoints
    # write txt for gridlabd to subscribe house setpoints and publish meter voltages

    dt = time_step
    period = 300 # not actually used
    mean_price = 0.1167
    std_dev_price = 0.0149
    k_slope = 1.0
    # autonomous precooling; if the meter voltage_1 exceeds vthresh, change the thermostat by toffset
    vthresh = 125.0
    toffset_min = -1.9
    toffset_max = -2.1

    gp = open (nameroot + '.glm', 'r')
    dp = open (nameroot + '_agent_dict.json', 'w')
    yp = open (nameroot + '_precool.yaml', 'w')
    cp = open (nameroot + '_FNCS_Config.txt', 'w')

    # write preambles
    print ('name: precool', file=yp)
    print ('time_delta: ' + str(dt) + 's', file=yp)
    print ('broker: tcp://localhost:5570', file=yp)
    print ('aggregate_sub: true', file=yp)
    print ('aggregate_pub: true', file=yp)
    print ('values:', file=yp)
    print ('  price:', file=yp)
    print ('    topic: player/price', file=yp)
    print ('    default:', mean_price, file=yp)

    # find the house and meter names
    inHouses = False
    endedHouse = False
    isELECTRIC = False
    houseName = ''
    meterName = ''
    houses = {}

    for line in gp:
        lst = line.split()
        if len(lst) > 1:
            if lst[1] == 'house':
                inHouses = True
            # Check for ANY object within the house, and don't use its name:
            if inHouses == True and lst[0] == 'object' and lst[1] != 'house':
                endedHouse = True
            # Check house object with controller inside
            if inHouses == True:
                if lst[0] == 'name' and endedHouse == False:
                    houseName = lst[1].strip(';')
                if lst[0] == 'parent':
                    meterName = lst[1].strip(';')
                if lst[0] == 'cooling_system_type':
                    if (lst[1].strip(';') == 'ELECTRIC'):
                        isELECTRIC = True
        elif len(lst) == 1:
            if inHouses == True: 
                inHouses = False
                endedHouse = False
                if isELECTRIC == True:
                    night_set = np.random.uniform (70, 76)
                    day_set = np.random.uniform (78, 82)
                    day_start = np.random.uniform (6, 8)
                    day_end = np.random.uniform (17, 19)
                    deadband = np.random.uniform (1, 2)
                    toffset = np.random.uniform (toffset_min, toffset_max)
                    houses[houseName] = {'meter':meterName,'night_set':float('{:.3f}'.format(night_set)),
                        'day_set':float('{:.3f}'.format(day_set)),'day_start_hour':float('{:.3f}'.format(day_start)),
                        'day_end_hour':float('{:.3f}'.format(day_end)),'deadband':float('{:.3f}'.format(deadband)),
                        'vthresh':vthresh,'toffset':toffset}
                    print ('  ' + houseName + '#V1:', file=yp)
                    print ('    topic: gld1/' + meterName + '/measured_voltage_1', file=yp)
                    print ('    default: 120', file=yp)
                    print ('  ' + houseName + '#Tair:', file=yp)
                    print ('    topic: gld1/' + houseName + '/air_temperature', file=yp)
                    print ('    default: 80', file=yp)
                    print ('publish \"commit:' + meterName + '.measured_voltage_1 -> ' + meterName + '/measured_voltage_1\";', file=cp)
                    print ('publish \"commit:' + houseName + '.air_temperature -> ' + houseName + '/air_temperature\";', file=cp)
                    print ('subscribe \"precommit:' + houseName + '.cooling_setpoint <- precool/' + houseName + '_cooling_setpoint\";', file=cp)
                    print ('subscribe \"precommit:' + houseName + '.heating_setpoint <- precool/' + houseName + '_heating_setpoint\";', file=cp)
                    print ('subscribe \"precommit:' + houseName + '.thermostat_deadband <- precool/' + houseName + '_thermostat_deadband\";', file=cp)
                    isELECTRIC = False

    meta = {'houses':houses,'period':period,'dt':dt,'mean':mean_price,'stddev':std_dev_price,'k_slope':k_slope}
    print (json.dumps(meta), file=dp)

    dp.close()
    gp.close()
    yp.close()
    cp.close()






#	Copyright (C) 2017 Battelle Memorial Institute
# file: process_houses.py; focus on HVAC
import json;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;

casefiles = [['SGIP1a','red'],
						 ['SGIP1b','blue'],
						 ['SGIP1c','green']]
#						 ['SGIP1d','magenta'],
#						 ['SGIP1e','cyan'],
#						 ['SGIP1ex','orange']]

def MakePlotData(root):
	lp = open (root + "_glm_dict.json")
	dict = json.loads(lp.read())
	lp.close()

	sub_keys = list(dict['feeders'].keys())
	sub_keys.sort()
	lp_s = open ("substation_" + root + "_metrics.json")
	lst_s = json.loads(lp_s.read())
	lp_s.close()
	lst_s.pop('StartTime')
	meta_s = lst_s.pop('Metadata')

	hse_keys = list(dict['houses'].keys())
	hse_keys.sort()
	for key in hse_keys:
		row = dict['houses'][key]
	lp_h = open ("house_" + root + "_metrics.json")
	lst_h = json.loads(lp_h.read())
	lp_h.close()
	lst_h.pop('StartTime')
	meta_h = lst_h.pop('Metadata')

	times = list(map(int,list(lst_h.keys())))
	times.sort()
	hrs = np.array(times, dtype=np.float)
	denom = 3600.0
	hrs /= denom

	time_key = str(times[0])

	for key, val in meta_s.items():
		if key == 'real_power_avg':
			SUB_POWER_IDX = val['index']

	for key, val in meta_h.items():
		if key == 'air_temperature_avg':
			AIR_AVG_IDX = val['index']
		elif key == 'hvac_load_avg':
			HVAC_AVG_IDX = val['index']

	data_s = np.empty(shape=(len(sub_keys), len(times), len(lst_s[time_key][sub_keys[0]])), dtype=np.float)
	j = 0
	for key in sub_keys:
		i = 0
		for t in times:
			ary = lst_s[str(t)][sub_keys[j]]
			data_s[j, i,:] = ary
			i = i + 1
		j = j + 1
	sub_mw = 1.0e-6 * data_s[0,:,SUB_POWER_IDX]

	data_h = np.empty(shape=(len(hse_keys), len(times), len(lst_h[time_key][hse_keys[0]])), dtype=np.float)
	j = 0
	for key in hse_keys:
		i = 0
		for t in times:
			ary = lst_h[str(t)][hse_keys[j]]
			data_h[j, i,:] = ary
			i = i + 1
		j = j + 1

	hvac1 = (data_h[:,:,HVAC_AVG_IDX]).squeeze()
	hvac2 = 0.001 * hvac1.sum(axis=0)

	j = 0
	n = 0
	avg_temp = np.zeros(len(times), dtype=np.float)
	for key in hse_keys:
		cool = dict['houses'][key]['cooling']
		if cool == 'ELECTRIC':
			avg_temp += data_h[j,:,AIR_AVG_IDX]
			n += 1
		j = j + 1
	avg_temp /= n
	print(n,'HVACs')
	return hrs, avg_temp, hvac2, sub_mw

def compare_hvac():
	# display a plot
	tmin = 0.0
	tmax = 48.0
	xticks = [0,6,12,18,24,30,36,42,48]

	fig, ax = plt.subplots(3, 1, sharex = 'col')

	for root in casefiles:
		print ('Processing', root[0])
		hrs, avg_temp, hvac2, sub_mw = MakePlotData(root[0])
		ax[0].plot(hrs, avg_temp, color=root[1], label=root[0])
		ax[1].plot(hrs, hvac2, color=root[1], label=root[0])
		ax[2].plot(hrs, sub_mw, color=root[1], label=root[0])

	ax[0].set_title ("Temperature at all HVAC Houses")
	ax[0].set_ylabel("Average Degrees")

	ax[1].set_title ("HVAC Power")
	ax[1].set_ylabel("Total MW")

	ax[2].set_title ("Feeder Power")
	ax[2].set_ylabel("Total MW")

	ax[2].set_xlabel("Hours")

	ax[0].grid()
	ax[0].legend()
	ax[0].set_xlim(tmin,tmax)
	ax[0].set_xticks(xticks)

	ax[1].grid()
	ax[1].legend()
	ax[1].set_xlim(tmin,tmax)
	ax[1].set_xticks(xticks)

	ax[2].grid()
	ax[2].legend()
	ax[2].set_xlim(tmin,tmax)
	ax[2].set_xticks(xticks)

	plt.show()





#	Copyright (C) 2017-2019 Battelle Memorial Institute
# file: process_pypower.py
"""Functions to plot bus and generator data from PYPOWER

Public Functions:
        :process_pypower: Reads the data and metadata, then makes the plots.  

"""
import json;
#import sys;
import numpy as np;
try:
  import matplotlib as mpl;
  import matplotlib.pyplot as plt;
except:
    pass

def process_pypower(nameroot):
    """ Plots bus and generator quantities for the 9-bus system used in te30 or sgip1 examples

    This function reads *bus_nameroot_metrics.json* and 
    *gen_nameroot_metrics.json* for the data, and 
    *nameroot_m_dict.json* for the metadata.  
    These must all exist in the current working directory.  
    One graph is generated with 8 subplots:

    1. Bus P and Q demands, at the single GridLAB-D connection
    2. Bus P and Q locational marginal prices (LMP), at the single GridLAB-D connection
    3. Bus Vmin, Vmax and Vavg, at the single GridLAB-D connection
    4. All 4 generator prices
    5. Generator 1 P and Q output 
    6. Generator 2 P and Q output 
    7. Generator 3 P and Q output 
    8. Generator 4 P and Q output 

    Args:
        nameroot (str): file name of the TESP case, not necessarily the same as the PYPOWER case, without the JSON extension
    """
    # first, read and print a dictionary of relevant PYPOWER objects
    lp = open (nameroot + "_m_dict.json").read()
    dict = json.loads(lp)
    baseMVA = dict['baseMVA']
    gen_keys = list(dict['generators'].keys())
    gen_keys.sort()
    bus_keys = list(dict['dsoBuses'].keys())
    bus_keys.sort()
    print ("\n\nFile", nameroot, "has baseMVA", baseMVA)
    print("\nGenerator Dictionary:")
    print("Unit Bus Type Pnom Pmax Costs[Start Stop C2 C1 C0]")
    for key in gen_keys:
        row = dict['generators'][key]
        print (key, row['bus'], row['bustype'], row['Pnom'], row['Pmax'], "[", row['StartupCost'], row['ShutdownCost'], row['c2'], row['c1'], row['c0'], "]")
    print("\nDSO Bus Dictionary:")
    print("Bus Pnom Qnom ampFactor [GridLAB-D Substations]")
    for key in bus_keys:
        row = dict['dsoBuses'][key]
        print (key, row['Pnom'], row['Qnom'], row['ampFactor'], row['GLDsubstations'])

    # read the bus metrics file
    lp_b = open ("bus_" + nameroot + "_metrics.json").read()
    lst_b = json.loads(lp_b)
    print ("\nBus Metrics data starting", lst_b['StartTime'])

    # make a sorted list of the times, and NumPy array of times in hours
    lst_b.pop('StartTime')
    #lst_b.pop('System base MVA')
    #lst_b.pop('Number of buses')
    #lst_b.pop('Number of generators')
    #lst_b.pop('Network name')
    meta_b = lst_b.pop('Metadata')
    times = list(map(int,list(lst_b.keys())))
    times.sort()
    print ("There are", len (times), "sample times at", times[1] - times[0], "second intervals")
    hrs = np.array(times, dtype=np.float)
    denom = 3600.0
    hrs /= denom

    # parse the metadata for things of specific interest
    print ("\nBus Metadata [Variable Index Units] for", len(lst_b[str(times[0])]), "objects")
    for key, val in meta_b.items():
    #    print (key, val['index'], val['units'])
        if key == 'LMP_P':
            LMP_P_IDX = val['index']
            LMP_P_UNITS = val['units']
        elif key == 'LMP_Q':
            LMP_Q_IDX = val['index']
            LMP_Q_UNITS = val['units']
        elif key == 'PD':
            PD_IDX = val['index']
            PD_UNITS = val['units']
        elif key == 'QD':
            QD_IDX = val['index']
            QD_UNITS = val['units']
        elif key == 'Vang':
            VANG_IDX = val['index']
            VANG_UNITS = val['units']
        elif key == 'Vmag':
            VMAG_IDX = val['index']
            VMAG_UNITS = val['units']
        elif key == 'Vmax':
            VMAX_IDX = val['index']
            VMAX_UNITS = val['units']
        elif key == 'Vmin':
            VMIN_IDX = val['index']
            VMIN_UNITS = val['units']

    # create a NumPy array of all bus metrics
    data_b = np.empty(shape=(len(bus_keys), len(times), len(lst_b[str(times[0])][bus_keys[0]])), dtype=np.float)
    print ("\nConstructed", data_b.shape, "NumPy array for Buses")
    j = 0
    for key in bus_keys:
        i = 0
        for t in times:
            ary = lst_b[str(t)][bus_keys[j]]
            data_b[j, i,:] = ary
            i = i + 1
        j = j + 1

    # display some averages
    print ("Average real power LMP = {:.5f} {:s}".format (data_b[0,:,LMP_P_IDX].mean(), LMP_P_UNITS))
    print ("Maximum real power LMP = {:.5f} {:s}".format (data_b[0,:,LMP_P_IDX].max(), LMP_P_UNITS))
    print ("First day LMP mean = {:.5f}".format (data_b[0,0:25,LMP_P_IDX].mean()))
    print ("First day LMP std dev = {:.6f}".format (data_b[0,0:25,LMP_P_IDX].std()))
    print ("Maximum bus voltage = {:.4f} {:s}".format ( data_b[0,:,VMAX_IDX].max(), VMAX_UNITS))
    print ("Minimum bus voltage = {:.4f} {:s}".format (data_b[0,:,VMIN_IDX].min(), VMIN_UNITS))

    # read the generator metrics file
    lp_g = open ("gen_" + nameroot + "_metrics.json").read()
    lst_g = json.loads(lp_g)
    print ("\nGenerator Metrics data starting", lst_g['StartTime'])
    # make a sorted list of the times, and NumPy array of times in hours
    lst_g.pop('StartTime')
    meta_g = lst_g.pop('Metadata')
    #print ("\nGenerator Metadata [Variable Index Units] for", len(lst_g[str(times[0])]), "objects")
    for key, val in meta_g.items():
        #print (key, val['index'], val['units'])
        if key == 'Pgen':
            PGEN_IDX = val['index']
            PGEN_UNITS = val['units']
        elif key == 'Qgen':
            QGEN_IDX = val['index']
            QGEN_UNITS = val['units']
        elif key == 'LMP_P':
            GENLMP_IDX = val['index']
            GENLMP_UNITS = val['units']

    # create a NumPy array of all bus metrics
    data_g = np.empty(shape=(len(gen_keys), len(times), len(lst_g[str(times[0])][gen_keys[0]])), dtype=np.float)
    print ("\nConstructed", data_g.shape, "NumPy array for Generators")
    j = 0
    for key in gen_keys:
        i = 0
        for t in times:
            ary = lst_g[str(t)][gen_keys[j]]
            data_g[j, i,:] = ary
            i = i + 1
        j = j + 1

    # display a plot - hard-wired assumption of 3 generators from Case 9
    fig, ax = plt.subplots(4,2, sharex = 'col')

    ax[0,0].plot(hrs, data_b[0,:,PD_IDX], color="blue", label="Real")
    ax[0,0].plot(hrs, data_b[0,:,QD_IDX], color="red", label="Reactive")
    ax[0,0].set_ylabel(PD_UNITS + "/" + QD_UNITS)
    ax[0,0].set_title ("Demands at " + bus_keys[0])
    ax[0,0].legend(loc='best')

    ax[1,0].plot(hrs, data_b[0,:,LMP_P_IDX], color="blue", label="Real")
    ax[1,0].plot(hrs, data_b[0,:,LMP_Q_IDX], color="red", label="Reactive")
    ax[1,0].set_ylabel(LMP_P_UNITS)
    ax[1,0].set_title ("Prices at " + bus_keys[0])
    ax[1,0].legend(loc='best')

    ax[2,0].plot(hrs, data_b[0,:,VMAG_IDX], color="blue", label="Magnitude")
    ax[2,0].plot(hrs, data_b[0,:,VMAX_IDX], color="red", label="Vmax")
    ax[2,0].plot(hrs, data_b[0,:,VMIN_IDX], color="green", label="Vmin")
    ax[2,0].set_ylabel(VMAG_UNITS)
    ax[2,0].set_title ("Voltages at " + bus_keys[0])
    ax[2,0].legend(loc='best')

    ax[3,0].plot(hrs, data_g[0,:,GENLMP_IDX], color='blue', label='unit 1')
    ax[3,0].plot(hrs, data_g[1,:,GENLMP_IDX], color='red', label='unit 2')
    ax[3,0].plot(hrs, data_g[2,:,GENLMP_IDX], color='green', label='unit 3')
    ax[3,0].plot(hrs, data_g[3,:,GENLMP_IDX], color='magenta', label='unit 4')
    ax[3,0].set_ylabel(GENLMP_UNITS)
    ax[3,0].set_title ('Generator Prices')
    ax[3,0].legend(loc='best')

    for i in range(0,4):
        ax[i,1].plot(hrs, data_g[i,:,PGEN_IDX], color="blue", label="P")
        ax[i,1].plot(hrs, data_g[i,:,QGEN_IDX], color="red", label="Q")
        ax[i,1].set_ylabel(PGEN_UNITS + "/" + QGEN_UNITS)
        ax[i,1].set_title ("Output from unit " + gen_keys[i])
        ax[i,1].legend(loc='best')

    ax[3,0].set_xlabel("Hours")
    ax[3,1].set_xlabel("Hours")

    plt.show()




#	Copyright (C) 2017 Battelle Memorial Institute
# file: process_houses.py; focus on HVAC
import json;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;

casefiles = [['SGIP1a','red'],
						 ['SGIP1b','blue'],
						 ['SGIP1c','green'],
						 ['SGIP1d','magenta'],
						 ['SGIP1e','cyan'],
						 ['SGIP1ex','orange']]

def MakePlotData(root):
	lp = open (root + "_m_dict.json")
	dict = json.loads(lp.read())
	lp.close()

	gen_keys = list(dict['generators'].keys())
	gen_keys.sort()
	lp_g = open ('gen_' + root + '_metrics.json')
	lst_g = json.loads(lp_g.read())
	lp_g.close()
	lst_g.pop('StartTime')
	meta_g = lst_g.pop('Metadata')

	bus_keys = list(dict['fncsBuses'].keys())
	bus_keys.sort()
	lp_b = open ('bus_' + root + '_metrics.json')
	lst_b = json.loads(lp_b.read())
	lp_b.close()
	lst_b.pop('StartTime')
	meta_b = lst_b.pop('Metadata')

	times = list(map(int,list(lst_b.keys())))
	times.sort()
	hrs = np.array(times, dtype=np.float)
	denom = 3600.0
	hrs /= denom

	time_key = str(times[0])

	for key, val in meta_g.items():
		if key == 'Pgen':
			PGEN_IDX = val['index']
			PGEN_UNITS = val['units']
		elif key == 'Qgen':
			QGEN_IDX = val['index']
			QGEN_UNITS = val['units']
		elif key == 'LMP_P':
			GENLMP_IDX = val['index']
			GENLMP_UNITS = val['units']

	for key, val in meta_b.items():
		if key == 'LMP_P':
			LMP_P_IDX = val['index']
			LMP_P_UNITS = val['units']
		elif key == 'LMP_Q':
			LMP_Q_IDX = val['index']
			LMP_Q_UNITS = val['units']
		elif key == 'PD':
			PD_IDX = val['index']
			PD_UNITS = val['units']
		elif key == 'QD':
			QD_IDX = val['index']
			QD_UNITS = val['units']
		elif key == 'Vang':
			VANG_IDX = val['index']
			VANG_UNITS = val['units']
		elif key == 'Vmag':
			VMAG_IDX = val['index']
			VMAG_UNITS = val['units']
		elif key == 'Vmax':
			VMAX_IDX = val['index']
			VMAX_UNITS = val['units']
		elif key == 'Vmin':
			VMIN_IDX = val['index']
			VMIN_UNITS = val['units']

	data_g = np.empty(shape=(len(gen_keys), len(times), len(lst_g[time_key][gen_keys[0]])), dtype=np.float)
	j = 0
	for key in gen_keys:
		i = 0
		for t in times:
			ary = lst_g[str(t)][gen_keys[j]]
			data_g[j, i,:] = ary
			i = i + 1
		j = j + 1

	data_b = np.empty(shape=(len(bus_keys), len(times), len(lst_b[time_key][bus_keys[0]])), dtype=np.float)
	j = 0
	for key in bus_keys:
		i = 0
		for t in times:
			ary = lst_b[str(t)][bus_keys[j]]
			data_b[j, i,:] = ary
			i = i + 1
		j = j + 1

	gen1 = (data_g[:,:,PGEN_IDX]).squeeze()
	tgen = gen1.sum(axis=0)
	pfncs = data_b[0,:,PD_IDX]
	lmp = data_b[0,:,LMP_P_IDX]
	return hrs, pfncs, tgen, lmp

def compare_pypower ():
	# display a plot
	tmin = 0.0
	tmax = 48.0
	xticks = [0,6,12,18,24,30,36,42,48]

	fig, ax = plt.subplots(3, 1, sharex = 'col')

	for root in casefiles:
		print ('Processing', root[0])
		hrs, pfncs, tgen, lmp = MakePlotData(root[0])
		ax[0].plot(hrs, pfncs, color=root[1], label=root[0])
		ax[1].plot(hrs, tgen, color=root[1], label=root[0])
		ax[2].plot(hrs, lmp, color=root[1], label=root[0])

	ax[0].set_title ('FNCS Bus Load')
	ax[0].set_ylabel ('MW')

	ax[1].set_title ('Total Generation')
	ax[1].set_ylabel ('MW')

	ax[2].set_title ('FNCS Bus LMP')
	ax[2].set_ylabel ('$/kwh')

	ax[2].set_xlabel ('Hours')

	ax[0].grid()
	ax[0].legend()
	ax[0].set_xlim(tmin,tmax)
	ax[0].set_xticks(xticks)

	ax[1].grid()
	ax[1].legend()
	ax[1].set_xlim(tmin,tmax)
	ax[1].set_xticks(xticks)

	ax[2].grid()
	ax[2].legend()
	ax[2].set_xlim(tmin,tmax)
	ax[2].set_xticks(xticks)

	plt.show()





# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: hvac.py
"""Class that controls the responsive thermostat for one house.

Implements the ramp bidding method, with HVAC power as the
bid quantity, and thermostat setting changes as the response
mechanism.
"""
import math
import tesp_support.helpers as helpers

class hvac:
    """This agent manages thermostat setpoint and bidding for a house

    Args:
        dict (dict): dictionary row for this agent from the JSON configuration file
        key (str): name of this agent, also key for its dictionary row
        aucObj (simple_auction): the auction this agent bids into

    Attributes:
        name (str): name of this agent
        control_mode (str): control mode from dict (CN_RAMP or CN_NONE, which still implements the setpoint schedule)
        houseName (str): name of the corresponding house in GridLAB-D, from dict
        meterName (str): name of the corresponding triplex_meter in GridLAB-D, from dict
        period (float): market clearing period, in seconds, from dict
        wakeup_start (float): hour of the day (0..24) for scheduled weekday wakeup period thermostat setpoint, from dict
        daylight_start (float): hour of the day (0..24) for scheduled weekday daytime period thermostat setpoint, from dict
        evening_start (float): hour of the day (0..24) for scheduled weekday evening (return home) period thermostat setpoint, from dict
        night_start (float): hour of the day (0..24) for scheduled weekday nighttime period thermostat setpoint, from dict
        wakeup_set (float): preferred thermostat setpoint for the weekday wakeup period, in deg F, from dict
        daylight_set (float): preferred thermostat setpoint for the weekday daytime period, in deg F, from dict
        evening_set (float): preferred thermostat setpoint for the weekday evening (return home) period, in deg F, from dict
        night_set (float): preferred thermostat setpoint for the weekday nighttime period, in deg F, from dict
        weekend_day_start (float): hour of the day (0..24) for scheduled weekend daytime period thermostat setpoint, from dict
        weekend_day_set (float): preferred thermostat setpoint for the weekend daytime period, in deg F, from dict
        weekend_night_start (float): hour of the day (0..24) for scheduled weekend nighttime period thermostat setpoint, from dict
        weekend_night_set (float): preferred thermostat setpoint for the weekend nighttime period, in deg F, from dict
        deadband (float): thermostat deadband in deg F, invariant, from dict
        offset_limit (float): maximum allowed change from the time-scheduled setpoint, in deg F, from dict
        ramp (float): bidding ramp denominator in multiples of the price standard deviation, from dict
        price_cap (float): the highest allowed bid price in $/kwh, from dict
        bid_delay (float): from dict, not implemented
        use_predictive_bidding (float): from dict, not implemented
        std_dev (float): standard deviation of expected price, determines the bidding ramp slope, initialized from aucObj
        mean (float): mean of the expected price, determines the bidding ramp origin, initialized from aucObj
        Trange (float): the allowed range of setpoint variation, bracketing the preferred time-scheduled setpoint
        air_temp (float): current air temperature of the house in deg F
        hvac_kw (float): most recent non-zero HVAC power in kW, this will be the bid quantity
        mtr_v (float): current line-neutral voltage at the triplex meter
        hvac_on (Boolean): True if the house HVAC is currently running
        basepoint (float): the preferred time-scheduled thermostat setpoint in deg F
        setpoint (float): the thermostat setpoint, including price response, in deg F
        bid_price (float): the current bid price in $/kwh
        cleared_price (float): the cleared market price in $/kwh
    """
    def __init__(self,dict,key,aucObj):
        """Initializes the class
        """
        self.name = key
        self.control_mode = dict['control_mode']
        self.houseName = dict['houseName']
        self.meterName = dict['meterName']
        self.period = float(dict['period'])
        self.wakeup_start = float(dict['wakeup_start'])
        self.daylight_start = float(dict['daylight_start'])
        self.evening_start = float(dict['evening_start'])
        self.night_start = float(dict['night_start'])
        self.wakeup_set = float(dict['wakeup_set'])
        self.daylight_set = float(dict['daylight_set'])
        self.evening_set = float(dict['evening_set'])
        self.night_set = float(dict['night_set'])
        self.weekend_day_start = float(dict['weekend_day_start'])
        self.weekend_day_set = float(dict['weekend_day_set'])
        self.weekend_night_start = float(dict['weekend_night_start'])
        self.weekend_night_set = float(dict['weekend_night_set'])
        self.deadband = float(dict['deadband'])
        self.offset_limit = float(dict['offset_limit'])
        self.ramp = float(dict['ramp'])
        self.price_cap = float(dict['price_cap'])
        self.bid_delay = float(dict['bid_delay'])
        self.use_predictive_bidding = float(dict['use_predictive_bidding'])

        self.std_dev = aucObj.std_dev
        self.mean = aucObj.clearing_price

        self.Trange = abs (2.0 * self.offset_limit)

        self.air_temp = 78.0
        self.hvac_kw = 3.0
        self.mtr_v = 120.0
        self.hvac_on = False

        self.basepoint = 0.0
        self.setpoint = 0.0
        self.cleared_price = 0.0
        self.bid_price = 0.0

    def inform_bid (self,price):
        """ Set the cleared_price attribute

        Args:
            price (float): cleared price in $/kwh
        """
        self.cleared_price = price

    def bid_accepted (self):
        """ Update the thermostat setting if the last bid was accepted

        The last bid is always "accepted". If it wasn't high enough,
        then the thermostat could be turned up.p

        Returns:
            Boolean: True if the thermostat setting changes, False if not.
        """
        if self.control_mode == 'CN_RAMP' and self.std_dev > 0.0:
            offset = (self.cleared_price - self.mean) * self.Trange / self.ramp / self.std_dev
            if offset < -self.offset_limit:
                offset = -self.offset_limit
            elif offset > self.offset_limit:
                offset = self.offset_limit
            self.setpoint = self.basepoint + offset
            return True
        return False

    def formulate_bid (self):
        """ Bid to run the air conditioner through the next period
        
        Returns:
            [float, float, Boolean]: bid price in $/kwh, bid quantity in kW and current HVAC on state, or None if not bidding 
        """

        #print (' = formulating bid for {:s} kw={:.2f} on={:d} T={:.2f} Base={:.2f} mu={:.5f} ramp={:.3f} std={:.5f} Trange={:.2f} mode={:s}'.format (self.name, 
        #  self.hvac_kw, self.hvac_on, self.air_temp, self.basepoint, self.mean, self.ramp, self.std_dev, self.Trange, self.control_mode))

        if self.control_mode == 'CN_NONE':
            return None

        p = self.mean + (self.air_temp - self.basepoint) * self.ramp * self.std_dev / self.Trange
        if p >= self.price_cap:
            self.bid_price = self.price_cap
        elif p <= 0.0:
            self.bid_price = 0.0
        else:
            self.bid_price = p
        return [self.bid_price, self.hvac_kw, self.hvac_on]

    def change_basepoint (self,hod,dow):
        """ Updates the time-scheduled thermostat setting

        Args:
            hod (float): the hour of the day, from 0 to 24
            dow (int): the day of the week, zero being Monday

        Returns:
            Boolean: True if the setting changed, Falso if not
        """
        if dow > 4: # a weekend
            val = self.weekend_night_set
            if hod >= self.weekend_day_start and hod < self.weekend_night_start:
                val = self.weekend_day_set
        else: # a weekday
            val = self.night_set
            if hod >= self.wakeup_start and hod < self.daylight_start:
                val = self.wakeup_set
            elif hod >= self.daylight_start and hod < self.evening_start:
                val = self.daylight_set
            elif hod >= self.evening_start and hod < self.night_start:
                val = self.evening_set
        if abs(self.basepoint - val) > 0.1:
            self.basepoint = val
            return True
        return False

    def set_hvac_load_from_fncs_str (self,str):
        """ Sets the hvac_load attribute, if greater than zero

        Args:
            str (str): FNCS message with load in kW
        """
        val = helpers.parse_fncs_number (str)
        if val > 0.0:
            self.hvac_kw = val

    def set_hvac_state_from_fncs_str (self,str):
        """ Sets the hvac_on attribute

        Args:
            str (str): FNCS message with state, ON or OFF
        """
        if str == 'OFF':
            self.hvac_on = False
        else:
            self.hvac_on = True

    def set_air_temp_from_fncs_str (self,str):
        """ Sets the air_temp attribute

        Args:
            str (str): FNCS message with temperature in degrees Fahrenheit
        """
        self.air_temp = helpers.parse_fncs_number (str)

    def set_voltage_from_fncs_str (self,str):
        """ Sets the mtr_v attribute

        Args:
            str (str): FNCS message with meter line-neutral voltage
        """
        self.mtr_v = helpers.parse_fncs_magnitude (str)




# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: fncs.py
""" Functions that provide access from Python to the FNCS library

Notes:
    Depending on the operating system, libfncs.dylib, libfncs.dll 
    or libfncs.so must already be installed. Besides the defined Python 
    wrapper functions, these pass-through library calls are always needed:

    - *fncs.finalize*: call after the simulation completes
    - *fncs.time_request (long long)*: request the next time step; blocks execution of this process until FNCS grants the requested time. Then, the process should check for messages from FNCS.

    These pass-through calls are also available, but not used in TESP:

    - *fncs.route*
    - *fncs.update_time_delta*
    - *fncs.get_id*
    - *fncs.get_simulator_count*
    - *fncs.get_events_size*
    - *fncs.get_keys_size*
    - *fncs.die*: stops FNCS and sends 'die' to other simulators

References:
    `ctypes <https://docs.python.org/3/library/ctypes.html>`_

    `FNCS <https://github.com/FNCS/fncs/>`_

Examples:
    - under tesp_support, see substation.py, precool.py and fncsPYPOWER.py
    - under examples, see loadshed/loadshed.py

"""
import ctypes
import platform

_libname = "libfncs.so"

if platform.system() == 'Darwin':
    _libname = "libfncs.dylib"
elif platform.system() == 'Windows':
    _libname = "libfncs"

try:
    _lib = ctypes.CDLL(_libname)

    _initialize = _lib.fncs_initialize
    _initialize.argtypes = []
    _initialize.restype = None

    _initialize_config = _lib.fncs_initialize_config
    _initialize_config.argtypes = [ctypes.c_char_p]
    _initialize_config.restype = None

    _agentRegister = _lib.fncs_agentRegister
    _agentRegister.argtypes = []
    _agentRegister.restype = None

    _agentRegisterConfig = _lib.fncs_agentRegisterConfig
    _agentRegisterConfig.argtypes = [ctypes.c_char_p]
    _agentRegisterConfig.restype = None

    _is_initialized = _lib.fncs_is_initialized
    _is_initialized.argtypes = []
    _is_initialized.restype = ctypes.c_int

    _time_request = _lib.fncs_time_request
    _time_request.argtypes = [ctypes.c_ulonglong]
    _time_request.restype = ctypes.c_ulonglong

    _publish = _lib.fncs_publish
    _publish.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
    _publish.restype = None

    _publish_anon = _lib.fncs_publish_anon
    _publish_anon.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
    _publish_anon.restype = None

    _agentPublish = _lib.fncs_agentPublish
    _agentPublish.argtypes = [ctypes.c_char_p]
    _agentPublish.restype = None

    _route = _lib.fncs_route
    _route.argtypes = [ctypes.c_char_p,
                      ctypes.c_char_p,
                      ctypes.c_char_p,
                      ctypes.c_char_p]
    _route.restype = None

    _die = _lib.fncs_die
    _die.argtypes = []
    _die.restype = None

    _finalize = _lib.fncs_finalize
    _finalize.argtypes = []
    _finalize.restype = None

    _update_time_delta = _lib.fncs_update_time_delta
    _update_time_delta.argtypes = [ctypes.c_ulonglong]
    _update_time_delta.restype = None

    _free = _lib._fncs_free
    _free.argtypes = [ctypes.c_void_p]
    _free.restype = None

    _get_events_size = _lib.fncs_get_events_size
    _get_events_size.argtypes = []
    _get_events_size.restype = ctypes.c_size_t

    _get_events = _lib.fncs_get_events
    _get_events.argtypes = []
    _get_events.restype = ctypes.POINTER(ctypes.POINTER(ctypes.c_char))

    _get_event_at = _lib.fncs_get_event_at
    _get_event_at.argtypes = [ctypes.c_size_t]
    _get_event_at.restype = ctypes.c_void_p

    _agentGetEvents = _lib.fncs_agentGetEvents
    _agentGetEvents.argtypes = []
    _agentGetEvents.restype = ctypes.c_void_p

    _get_value = _lib.fncs_get_value
    _get_value.argtypes = [ctypes.c_char_p]
    _get_value.restype = ctypes.c_void_p

    _get_values_size = _lib.fncs_get_values_size
    _get_values_size.argtypes = [ctypes.c_char_p]
    _get_values_size.restype = ctypes.c_size_t

    _get_values = _lib.fncs_get_values
    _get_values.argtypes = [ctypes.c_char_p]
    _get_values.restype = ctypes.POINTER(ctypes.POINTER(ctypes.c_char))

    _get_value_at = _lib.fncs_get_value_at
    _get_value_at.argtypes = [ctypes.c_char_p, ctypes.c_size_t]
    _get_value_at.restype = ctypes.c_void_p

    _get_keys_size = _lib.fncs_get_keys_size
    _get_keys_size.argtypes = []
    _get_keys_size.restype = ctypes.c_size_t

    _get_keys = _lib.fncs_get_keys
    _get_keys.argtypes = []
    _get_keys.restype = ctypes.POINTER(ctypes.POINTER(ctypes.c_char))

    _get_key_at = _lib.fncs_get_key_at
    _get_key_at.argtypes = [ctypes.c_size_t]
    _get_key_at.restype = ctypes.c_void_p

    _get_name = _lib.fncs_get_name
    _get_name.argtypes = []
    _get_name.restype = ctypes.c_void_p

    _get_id = _lib.fncs_get_id
    _get_id.argtypes = []
    _get_id.restype = ctypes.c_int

    _get_simulator_count = _lib.fncs_get_simulator_count
    _get_simulator_count.argtypes = []
    _get_simulator_count.restype = ctypes.c_int

    _get_version = _lib.fncs_get_version
    _get_version.argtypes = [ctypes.POINTER(ctypes.c_int),
                             ctypes.POINTER(ctypes.c_int),
                             ctypes.POINTER(ctypes.c_int)]
    _get_version.restype = None
except:
    pass

def time_request(time):
    """ FNCS time request

    Args:
        time (int): requested time.
    """
    return _time_request(time)

def initialize(config=None):
    """ Initialize the FNCS configuration

    Args:
        config (str): a ZPL file. If None (default), provide YAML file in FNCS_CONFIG_FILE environment variable.
    """
    if config:
        _initialize_config(config)
    else:
        _initialize()

def agentRegister(config=None):
    """ Initialize the FNCS configuration for the agent interface

    Args:
        config (str): a ZPL file. If None (default), provide YAML file in FNCS_CONFIG_FILE environment variable.
    """
    if config:
        _agentRegisterConfig(config)
    else:
        _agentRegister()

def is_initialized():
    """ Determine whether the FNCS library has been initialized

    Returns:
        Boolean: True if initialized, False if not.
    """
    return 1 == _is_initialized()

def publish(key, value):
    """ Publish a value over FNCS, under the simulator name

    Args:
        key (str): topic under the simulator name
        value (str): value
    """
    _publish(str(key).encode('utf-8'), str(value).encode('utf-8'))

def publish_anon(key, value):
    """ Publish a value over FNCS, under the 'anonymous' simulator name

    Args:
        key (str): topic under 'anonymous'
        value (str): value
    """
    _publish_anon(str(key).encode('utf-8'), str(value).encode('utf-8'))

def agentPublish(value):
    """ Publish a value over FNCS, under the configured simulator name / agent name

    Args:
        value (str): value
    """
    _agentPublish(str(value).encode('utf-8'))

def route(sender, receiver, key, value):
    """ Route a value over FNCS from sender to receiver

    Args:
        sender (str): simulator routing the message
        receiver (str): simulator to route the message to
        key (str): topic under the simulator name
        value (str): value
    """
    _route(str(sender).encode('utf-8'), str(receiver).encode('utf-8'), str(key).encode('utf-8'), str(value).encode('utf-8'))

def get_events():
    """ Retrieve FNCS messages after time_request returns

    Returns:
        list: tuple of decoded FNCS events
    """
    _events = _get_events()
    size = get_events_size()
    events_tmp = [ctypes.cast(_events[i], ctypes.c_char_p).value for i in range(size)]
    events = [x.decode() for x in events_tmp]
    for i in range(size):
        _free(_events[i])
    _free(_events)
    return events

def get_event_at(i):
    """ Retrieve FNCS message by index number after time_request returns

    Returns:
        str: one decoded FNCS event
    """
    _event = _get_event_at(i)
    event_tmp = ctypes.string_at(ctypes.cast(_event, ctypes.c_char_p).value)
    event = event_tmp.decode()
    _free(_event)
    return event

def agentGetEvents():
    """ Retrieve FNCS agent messages

    Returns:
        str: concatenation of agent messages
    """
    _event = _agentGetEvents()
    event_tmp = ctypes.string_at(ctypes.cast(_event, ctypes.c_char_p).value)
    event = event_tmp.decode()
    _free(_event)
    return event

def get_value(key):
    """ Extract value from a FNCS message

    Args:
        key (str): the topic

    Returns:
        str: decoded value
    """
    _value = _get_value(key.encode('utf-8'))
    value_tmp = ctypes.string_at(ctypes.cast(_value, ctypes.c_char_p).value)
    value = value_tmp.decode()
    _free(_value)
    return value

def get_values_size(key):
    """ For list publications, find how many values were published

    Args:
        key (str): the topic

    Returns:
        int: the number of values for this topic
    """
    return _get_values_size(key.encode('utf-8'))

def get_values(key):
    """ For list publications, get the list of values

    Args:
        key (str): the topic

    Returns:
        [str]: decoded values
    """
    _key = key.encode('utf-8')
    _values = _get_values(_key)
    size = get_values_size(_key)
    values_tmp = [ctypes.cast(_values[i], ctypes.c_char_p).value for i in range(size)]
    values = [x.decode() for x in values_tmp]
    for i in range(size):
        _free(_values[i])
    _free(_values)
    return values

def get_value_at(key, i):
    """ For list publications, get the value by index

    Args:
        key (str): the topic
        i (int): the list index number

    Returns:
        str: decoded value
    """
    _value = _get_value_at(key.encode('utf-8'), i)
    value_tmp = ctypes.string_at(ctypes.cast(_value, ctypes.c_char_p).value)
    value = value_tmp.decode()
    _free(_value)
    return value

def get_keys():
    """ Find the list of topics

    Returns:
        [str]: decoded topic names
    """
    _keys = _get_keys()
    size = get_keys_size()
    keys_tmp = [ctypes.cast(_keys[i], ctypes.c_char_p).value for i in range(size)]
    keys = [x.decode() for x in keys_tmp]
    for i in range(size):
        _free(_keys[i])
    _free(_keys)
    return keys

def get_key_at(i):
    """ Get the topic by index number

    Args:
        i (int): the index number

    Returns:
        str: decoded topic name
    """
    _key = _get_key_at(i)
    key_tmp = ctypes.string_at(ctypes.cast(_key, ctypes.c_char_p).value)
    key = key_tmp.decode()
    _free(_key)
    return key

def get_name():
    """ Find the FNCS simulator name

    Returns:
        str: the name of this simulator as provided in the ZPL or YAML file
    """
    _name = _get_name()
    name_tmp = ctypes.string_at(ctypes.cast(_name, ctypes.c_char_p).value)
    name = name_tmp.decode()
    _free(_name)
    return name

def get_version():
    """ Find the FNCS version

    Returns:
        int, int, int: major, minor and patch numbers
    """
    major = ctypes.c_int()
    minor = ctypes.c_int()
    patch = ctypes.c_int()
    _get_version(ctypes.byref(major),
                 ctypes.byref(minor),
                 ctypes.byref(patch))
    return (major.value, minor.value, patch.value)

def die():
    """ Call FNCS die because of simulator error

    """
    _die()

def finalize():
    """ Call FNCS finalize to end connection with broker

    """
    _finalize()

def update_time_delta(delta):
    """ Update simulator time delta value

    Args:
        delta (int): time delta.
    """
    _update_time_delta(delta)

def get_events_size():
    """ Get the size of the event queue

    """
    return _get_events_size()

def get_keys_size():
    """ Get the size of the keys

    """
    return _get_keys_size()

def get_simulator_count():
    """ Find the FNCS simulator count

    """
    return _get_simulator_count()

def get_id():
    """ Find the FNCS ID

    """
    return _get_id()



"""Weather Agent

This weather agent needs an WEATHER_CONFIG environment variable to be set, which is a json file.
"""
import sys, os
import pandas as pd
import json
from datetime import datetime
from datetime import timedelta
import random
import numpy
from scipy.stats import truncnorm
import time
try:
    import helics  # set the broker = HELICS in WEATHER_CONFIG
except:
    pass

def stop_helics_federate (fed):
  status = helics.helicsFederateFinalize(fed)
  state = helics.helicsFederateGetState(fed)
  assert state == 3
  while helics.helicsBrokerIsConnected(None):
    time.sleep(1)
  helics.helicsFederateFree(fed)
  helics.helicsCloseLibrary()

def startWeatherAgent(file):
    """the weather agent publishes weather data as configured by the json file

    :param file: string
        the weather data file
    :return: nothing
    """
    # read the weather data file, arguments to mimic deprecated from_csv function
    weatherData = pd.read_csv(file, index_col=0, parse_dates=True)
    config = os.environ['WEATHER_CONFIG'] # read the weather config json file
    if os.path.isfile(config):
        with open(config, 'r') as stream:
            try:
                conf = json.load(stream)
                agentName = conf['name']
                broker = conf['broker']
                timeStop = conf['time_stop']
                StartTime = conf['StartTime']
                timeFormat = '%Y-%m-%d %H:%M:%S'
                dtStart = datetime.strptime (StartTime, timeFormat)
                timeDeltaStr = conf['time_delta']
                publishInterval = conf['publishInterval']
                forecast = conf['Forecast']
                addErrorToForecast = conf['AddErrorToForecast']
                forecastLength = conf['ForecastLength']
                publishTimeAhead = conf['PublishTimeAhead']
                forecastPeriod = conf['forecastPeriod']
                forecastParameters = conf['parameters']
            except ValueError as ex:
                print(ex)
    else:
        print('could not open WEATHER_CONFIG file for FNCS or HELICS')
        sys.exit()

    # convert some of the time in config file to seconds
    try:
        publishTimeAhead = convertTimeToSeconds(publishTimeAhead)
    except Exception as ex:
        print("Error in PublishTimeAhead", ex)

    try:
        timeDeltaInSeconds = convertTimeToSeconds(timeDeltaStr)
    except Exception as ex:
        print("Error in time_delta", ex)

    try:
        publishIntervalInSeconds = convertTimeToSeconds(publishInterval)
    except Exception as ex:
        print("Error in publish Interval", ex)

    try:
        forecastLength = convertTimeToSeconds(forecastLength)
    except Exception as ex:
        print("Error in ForecastLength", ex)

    try:
        timeStopInSeconds = convertTimeToSeconds(timeStop)
    except Exception as ex:
        print("Error in time_stop", ex)

    #write fncs.zpl file here
    #it actually worked now. # this config str won't work as an argument to fncs::initialize, so write fncs.zpl just in time
    zplstr = "name = {}\ntime_delta = {}s\ntime_stop = {}s\nbroker = {}".format(agentName, timeDeltaInSeconds, timeStopInSeconds, broker)

    # when doing resample(), use publishIntervalInSeconds to make it uniform
    # the reason for that is due to some of the units that we use for fncs, such as 'min',
    # is not recognized by the resample() function
    weatherData2 = weatherData.resample(rule=str(publishIntervalInSeconds)+"s",closed='left').first()
    weatherData2 = weatherData2.interpolate(method='quadratic')

    # find weather data on the hour for the hourly forecast
    hourlyWeatherData=weatherData.loc[(weatherData.index.minute == 0) & (weatherData.index.second == 0) & (weatherData.index.microsecond == 0) & (weatherData.index.nanosecond == 0)]

    # find all the time point that the data at that time need to be published
    timeNeedToPublishRealtime = [0]
    timeNeedToPublishForecast = [0]
    # real time need to publish
    numberOfRealtimeBroadcast = timeStopInSeconds // publishIntervalInSeconds + 1
    for i in range(1,numberOfRealtimeBroadcast):
        timeNeedToPublishRealtime.append(i * publishIntervalInSeconds)
    if forecast == 1:
        # time need to publish forecast, which is on the hour
        numberOfForecast = timeStopInSeconds // 3600 + 1
        for i in range(1,numberOfForecast):
            timeNeedToPublishForecast.append(i * 3600)
        # combine real time and forecast time
        timeNeedToBePublished = list(set([0] + timeNeedToPublishRealtime + timeNeedToPublishForecast))
    else:
        timeNeedToBePublished = timeNeedToPublishRealtime
    timeNeedToBePublished.sort()

    # find all the time point that need to publish weather data,
    # each time point in this list pairs with each time point in timeNeedToBePublished list
    timeNeedToPublish = [(i - publishTimeAhead) if (i - publishTimeAhead) >= 0 else 0 for i in timeNeedToBePublished]

    # other weather agents could be initializing from FNCS.zpl, so we might have a race condition
    #  file locking didn't work, because fncs.initialize() doesn't return until broker hears from all other simulators
    hFed = None
    hPubs = {}
    fedName = agentName # 'weather'
    if broker == 'HELICS':
      fedInfo = helics.helicsCreateFederateInfo()
      helics.helicsFederateInfoSetCoreName(fedInfo, fedName)
      helics.helicsFederateInfoSetCoreTypeFromString(fedInfo, 'zmq')
      helics.helicsFederateInfoSetCoreInitString(fedInfo, '--federates=1')
      helics.helicsFederateInfoSetTimeProperty(fedInfo, helics.helics_property_time_delta, timeDeltaInSeconds)
      hFed = helics.helicsCreateValueFederate(fedName, fedInfo)
      for col in weatherData.columns:
        pubName = fedName + '/' + col
        hPubs[col] = helics.helicsFederateRegisterGlobalPublication(hFed, pubName, helics.helics_data_type_string, "")
        pubName = pubName + '/forecast'
        hPubs[col + '/forecast'] = helics.helicsFederateRegisterGlobalPublication(hFed, pubName, helics.helics_data_type_string, "")
      helics.helicsFederateEnterExecutingMode(hFed)
      print('HELICS initialized to publish', hPubs, flush=True)
    else:
      try:
        import tesp_support.fncs as fncs
      except:
        pass
      configstr = zplstr.encode('utf-8')
      fncs.initialize(configstr)
      print('FNCS initialized', flush=True)

    time_granted = 0
    for i in range(len(timeNeedToPublish)):
        if i > 0:
            timeToRequest = timeNeedToPublish[i]
            if hFed is not None:
              time_granted = int (helics.helicsFederateRequestTime(hFed, timeToRequest))
            else:
              time_granted = fncs.time_request(timeToRequest)
        if timeNeedToBePublished[i] in timeNeedToPublishRealtime:
            # find the data by the time point and publish them
            row = weatherData2.loc[dtStart + timedelta(seconds=timeNeedToBePublished[i])]
            print('publishing at ' + str(dtStart + timedelta(seconds=timeNeedToPublish[i]))
                  + ' for weather at ' + str(dtStart + timedelta(seconds=timeNeedToBePublished[i])), flush=True)
            for key, value in row.iteritems():
                #remove the improper value generated by interpolation
                if key != "temperature" and value < 1e-4:
                    value = 0
                if hFed is not None:
                  helics.helicsPublicationPublishDouble(hPubs[key], value)
                else:
                  fncs.publish(key, value)
        # if forecasting needed and the time is on the hour
        if forecast == 1 and timeNeedToBePublished[i] in timeNeedToPublishForecast:
            print('forecasting at ' + str(dtStart + timedelta(seconds=timeNeedToPublish[i])) + ' for weather starting from '
                  + str(dtStart + timedelta(seconds=timeNeedToBePublished[i])), flush=True)
            forecastStart = dtStart + timedelta(seconds=timeNeedToBePublished[i])
            forecastEnd = dtStart + timedelta(seconds=forecastLength) + timedelta(seconds=timeNeedToBePublished[i])
            # find the data by forecast starting and ending time, should be multiple data point for each weather factor
            rows = hourlyWeatherData.loc[(hourlyWeatherData.index >= forecastStart) & (hourlyWeatherData.index < forecastEnd)].copy()
            rows.solar_direct[rows.solar_direct<1e-4]=0
            rows.solar_diffuse[rows.solar_diffuse<1e-4]=0
            rows.wind_speed[rows.wind_speed<1e-4]=0
            rows.humidity[rows.humidity<1e-4]=0
            rows.pressure[rows.pressure<1e-4]=0
            for col in rows.columns:
                data = rows[col].values
                times = rows.index
                # if user wants to add error to the forecasted data to mimick weather forecast
                if addErrorToForecast == 1:
                    WF_obj = weather_forecast(col, forecastPeriod * 2, forecastParameters)  # make object
                    data = WF_obj.make_forecast(data, len(data))
                wd = dict()
                # convert data to a dictionary with time as the key so it can be published as json string
                for v in range(len(data)):
                    if col != "temperature" and data[v] < 1e-4:
                        data[v] = 0
                    wd[str(times[v])] = str(data[v])
                if hFed is not None:
                  helics.helicsPublicationPublishString(hPubs[col + '/forecast'], json.dumps(wd))
                else:
                  fncs.publish(col + '/forecast', json.dumps(wd))

    # if the last time step/stop time is not requested
    if timeStopInSeconds not in timeNeedToPublish:
        if hFed is not None:
          time_granted = int (helics.helicsFederateRequestTime(hFed, timeStopInSeconds))
        else:
          time_granted = fncs.time_request(timeStopInSeconds)

    if hFed is not None:
      stop_helics_federate (hFed)
    else:
      print('finalizing FNCS', flush=True)
      fncs.finalize()

def usage():
    print("usage: python weatherAgent.py <input weather file full path>")

def convertTimeToSeconds(time):
    """Convert time string with unit to integer in seconds

    It only parse unit in day, hour, minute and second.
    It will not recognize week, month, year, millisecond, microsecond or nanosecond, they can be added if needed.

    :param time: str
        time with unit
    :return: int
        represent the input time in second
    """
    unit = ''.join(filter(str.isalpha, time))
    timeNum = int(''.join(filter(str.isdigit, time)))
    if ("d" == unit or "day" == unit or "days" == unit):
        return 24 * 60 * 60 * timeNum
    elif ("h" == unit or "hour" == unit or "hours" == unit):
        return 60 * 60 * timeNum
    elif ("m" == unit or "min" == unit or "minute" == unit or "minutes" == unit):
        return 60 * timeNum
    elif ("s" == unit or "sec" == unit or "second" == unit or "seconds" == unit):
        return timeNum
    else:
        raise Exception("unrecognized time unit '" + unit + "'.")

def deltaTimeToResmapleFreq(time):
    """Convert time unit to a resampling frequency that can be recognized by pandas.DataFrame.resample()

    It only parse unit in day, hour, minute and second.
    It won't recognize week, month, year, millisecond, microsecond or nanosecond, they can be added if needed.

    :param time: str
        time with unit
    :return: str
        time with resample frequency
    """
    unit = ''.join(filter(str.isalpha, time))
    timeNum = int(''.join(filter(str.isdigit, time)))
    if ("d" == unit or "day" == unit or "days" == unit):
        return str(timeNum) + "d"
    elif ("h" == unit or "hour" == unit or "hours" == unit):
        return str(timeNum) + "h"
    elif ("m" == unit or "min" == unit or "minute" == unit or "minutes" == unit):
        return str(timeNum) + "T"
    elif ("s" == unit or "sec" == unit or "second" == unit or "seconds" == unit):
        return str(timeNum) + "s"
    else:
        raise Exception("unrecognized time unit '" + unit + "'.")

def findDeltaTimeMultiplier(time):
    """find the multiplier to convert delta_time to seconds

    It only parse unit in day, hour, minute and second.
    It won't recognize week, month, year, millisecond, microsecond or nanosecond, they can be added if needed.

    :param time: str
        time with unit
    :return: int
        the multiplier to convert delta_time to seconds
    """
    unit = ''.join(filter(str.isalpha, time))
    timeNum = int(''.join(filter(str.isdigit, time)))
    if ("d" == unit or "day" == unit or "days" == unit):
        return 24 * 60 * 60
    elif ("h" == unit or "hour" == unit or "hours" == unit):
        return 60 * 60
    elif ("m" == unit or "min" == unit or "minute" == unit or "minutes" == unit):
        return 60
    elif ("s" == unit or "sec" == unit or "second" == unit or "seconds" == unit):
        return 1
    else:
        raise Exception("unrecognized time unit '" + unit + "'.")

"""Class that includes error to the known Weather data 

Implements the range of values the errors are randomly selected. The range is time
dependent, i.e., the next hour range of errors are smaller than other error ranges.
The error range is referred to as "envelope" (error envelope of possible variables).
The envelope increases linearly.

The class also possesses a time varying bias which will displace one side of the 
envelope. However, only one side, thus not changing the maximum possible errors.

All the variables utilize in the class are time dependent. Thus, arrays where 
element "0" is the next hour and so forth.   

"""
class weather_forecast:
    """This object includes the error to a weather variable

    Args:
        variable (str): Type of weather variable being forecasted
        period (int): period of the sinusoidal bias
        W_dict (dict): dictionary for specifying the generation of the error envelope

    Attributes:
        weather_variable (str): Type of weather variable being forecasted
        # Type of error insertion
        distribution (int): type of distribution --> 0 uniform;1 triangular;2 truncated normal the standard deviation is computed for 95% of values to be within bounds in a conventional normal distribution
        P_e_bias (float): pu maximum bias at first hour --> [0 to 1]
        P_e_envelope (float): pu maximum error from mean values --> [0 to 1]
        Lower_e_bound (float): pu of the maximum error at the first hour --> [0 to 1]
        # Bias variable
        biasM (float) (1 X period): sinusoidal bias for altering the error envelope
        Period_bias (int): period of the sinusoidal bias
    """

    def __init__(self, variable, period, W_dict):
        """Initializes the class
        """
        self.weather_variable = variable
        self.Period_bias = period
        ############## Including a bias to the envelope
        # sinusoidal with a period of two times the size of y
        self.biasM = numpy.sin(numpy.linspace(-numpy.pi, numpy.pi, (period + 1)))
        self.biasM = self.biasM[:-1]
        self.forecastParameters = W_dict
        self.distribution = W_dict[variable]["distribution"]
        self.P_e_bias = W_dict[variable]["P_e_bias"]
        self.P_e_envelope = W_dict[variable]["P_e_envelope"]
        self.Lower_e_bound = W_dict[variable]["Lower_e_bound"]

    def get_truncated_normal(self, EL, EH):
        """Truncated normal distribution
        """
        mean=(EL+EH)/2
        sd=(abs(EL)+abs(EH))/4 #95% of values are within bounds remaining is truncated
        if sd <= 0.0:
            return 0.0
        a = (EL - mean) / sd
        b = (EH - mean) / sd
        sample = truncnorm.rvs(a,b,loc=mean,scale=sd,size=1)[0]
        return sample

    def make_forecast(self, weather, t=0):
        """Include error to a known weather variable

        Args:
            weather (float) (1 x desired number of hours ahead): known weather variable
            t (int): time in hours

        Returns:
            weather_f (float) (1 x desired number of hours ahead): weather variable with included error
            ENV_U (float) (1 x desired number of hours ahead): envelope with bias upper bound
            ENV_l (float) (1 x desired number of hours ahead): envelope with bias lower bound

        """
        ############## Making the error envelope
        scale = numpy.linspace(self.Lower_e_bound, 1, num=len(weather))  # error increases true time
        envelope = scale * numpy.mean(weather) * self.P_e_envelope
        ############## Including a bias to the envelope
        bias = self.biasM * (min(envelope) * 2 * self.P_e_bias)
        bias = numpy.roll(bias, -t)
        ############## making the error array
        n = len(weather)
        error = numpy.zeros(n)
        ############## sampling the error distribution
        ENV_l = list()
        ENV_U = list()
        for i in range(n):
            if bias[i] > 0:
                EL = -envelope[i] + bias[i]
                EH = envelope[i]
            else:
                EL = -envelope[i]
                EH = envelope[i] + bias[i]
            ENV_l.append(EL)
            ENV_U.append(EH)
            if self.distribution == 0:  # uniform
                error[i] = random.uniform(EL, EH)
            elif self.distribution == 1:  # triangular
                error[i] = random.triangular(EL, EH)
            elif self.distribution == 2:  # truncated normal 95%
                error[i] = self.get_truncated_normal(EL, EH)

        weather_f = error + weather
        return weather_f

if __name__ == '__main__':
    if len(sys.argv) != 2:
        usage()
        sys.exit()
    inputFile = sys.argv[1]
    startWeatherAgent(inputFile)



# Copyright (C) 2020 Battelle Memorial Institute
# file: make_ems.py
"""Creates and merges the EMS for an EnergyPlus building model

Public Functions:
  :make_ems: Creates the energy management system (EMS) for FNCS/HELICS to interface with EnergyPlus.
  :merge_idf: Assembles the base IDF, the EMS, start time and end time
"""
import sys
import re
import csv
from datetime import datetime
import tesp_support.helpers as helpers

if sys.platform == 'win32':
  pycall = 'python'
else:
  pycall = 'python3'

def valid_var (name):
  return name.replace (' ', '_').replace ('-', '_')

def schedule_sensor(name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s},  !- Name'.format (name), file=op)
  print ('    {:s},  !- Output:Variable or Output:Meter Index Key Name'.format (name), file=op)
  print ('    Schedule Value;    !- Output:Variable or Output:Meter Name', file=op)

def schedule_actuator(name, target, op):
  print ('  EnergyManagementSystem:Actuator,', file=op)
  print ('    {:s},  !- Name'.format (name), file=op)
  print ('    {:s},  !- Actuated Component Unique Name'.format (target), file=op)
  print ('    Schedule:Compact, !- Actuated Component Type', file=op)
  print ('    Schedule Value;   !- Actuated Component Control Type', file=op)

def global_variable(name, op):
  print ('  EnergyManagementSystem:GlobalVariable,', file=op)
  print ('    {:s};'.format (name), file=op)

def output_variable(name, target, op):
  print ('  EnergyManagementSystem:OutputVariable,', file=op)
  print ('    {:s},  !- Name'.format (name), file=op)
  print ('    {:s},  !- EMS Variable Name'.format (target), file=op)
  print ('    Averaged,     !- Type of Data in Variable', file=op)
  print ('    ZoneTimeStep, !- Update Frequency', file=op)
  print ('    ,             !- EMS Program or Subroutine Name', file=op)
  print ('    ;             !- Units', file=op)

def heating_coil_sensor(name, target, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s},  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},  !- Coil'.format (target), file=op)
  print ('    Heating Coil Electric Energy;', file=op)

def cooling_coil_sensor(name, target, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s},  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},  !- Coil'.format (target), file=op)
  print ('    Cooling Coil Electric Energy;', file=op)

def zone_temperature_sensor(name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s}_T,  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},    !- Zone'.format (name), file=op)
  print ('    Zone Mean Air Temperature;', file=op)

def zone_heating_sensor(name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s}_H,                   !- Name'.format (valid_var(name)), file=op)
  print ('    {:s} VAV Box Reheat Coil, !- Zone/Coil'.format (name), file=op)
  print ('    Heating Coil Heating Energy;', file=op)

def zone_sensible_heating_sensor(name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s}_H,  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},    !- Zone'.format (name), file=op)
  print ('    Zone Air System Sensible Heating Energy;', file=op)

def zone_sensible_cooling_sensor(name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s}_C,  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},    !- Zone'.format (name), file=op)
  print ('    Zone Air System Sensible Cooling Energy;', file=op)

def zone_occupant_sensor (name, op):
  print ('  EnergyManagementSystem:Sensor,', file=op)
  print ('    {:s}_O,  !- Name'.format (valid_var(name)), file=op)
  print ('    {:s},    !- Zone'.format (name), file=op)
  print ('    Zone People Occupant Count;', file=op)

def get_eplus_token (sval):
  val = sval.strip()
  idx = val.rfind(';')
  if idx < 0:
    idx = val.rfind(',')
  return val[:idx].upper()

def print_idf_summary (target, zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs):
  print ('  === hvacs', hvacs)
  print ('\n  === ccoils                             Sensor')
  for name, row in ccoils.items():
    print ('{:40s} {:s}'.format (name, row['Sensor']))
  print ('\n  === hcoils                             Sensor')
  for name, row in hcoils.items():
    print ('{:40s} {:s}'.format (name, row['Sensor']))
  print ('\n  === schedules used                     Alias      Heating')
  for name, row in schedules.items():
    if row['Used']:
      print ('{:40s} {:10s} {:1}'.format(name, row['Alias'], row['Heating']))
      print (row['Schedule'])
  print ('\n  === thermostats                        Heating                                  Cooling')
  for name, row in thermostats.items():
    print ('{:40s} {:40s} {:40s}'.format (name, row['Heating'], row['Cooling']))
  print ('\n  === zonecontrols                       Thermostat')
  for name, row in zonecontrols.items():
    print ('{:40s} {:40s}'.format (name, row))
  print ('\n  === zones                                Volume   Heating                                    Cooling                                  People Controlled')
  for zname, row in zones.items():
    zvol = row['zvol']
    Hsched = row['Hsched']
    Csched = row['Csched']
    People = row['People']
    Controlled = row['Controlled']
    print ('{:40s} {:8.2f}   {:40s}   {:40s} {:1}      {:1}'.format (zname, zvol, Hsched, Csched, People, Controlled))

def summarize_idf (fname, baseidf):
  schedules = {}
  thermostats = {}
  zonecontrols = {}
  zones = {}
  hvacs = set()
  ccoils = {}
  hcoils = {}
  volume = 0
  nzones = 0
  ncontrols = 0
  nsetpoints = 0
  nschedused = 0
  idx_ccoil = 1
  idx_hcoil = 1

  fp = open(fname, 'r')
  rdr = csv.reader (fp)
  for row in rdr:
    if row[0].strip() == 'Component Sizing Information':
      if row[1].strip() == 'AirLoopHVAC':
        HVACname = row[2].strip()
        hvacs.add (HVACname)
    if row[0].strip() == 'Zone Information':
      zname = row[1].strip()
      zvol = float(row[19])
      nzones += 1
      volume += zvol
      zones[zname] = {'zvol': zvol, 'Hsched': '', 'Csched': '', 'People': False, 'Controlled': False}
  fp.close()

  fp = open(baseidf, 'r', errors='replace')
  line = fp.readline()
  while line:
    if '!' not in line[0]:
      if 'People,' in line:
        name = get_eplus_token (fp.readline())
        zname = get_eplus_token (fp.readline())
        zones[zname]['People'] = True
        while ';' not in line:
          line = fp.readline()
      if ('Coil:Heating:Electric' in line) or ('Coil:Heating:DX' in line):
        coilname = get_eplus_token (fp.readline())
        if coilname not in hcoils:
          hcoils[coilname] = {'Sensor': 'Heating_Coil_{:d}'.format(idx_hcoil)}
          idx_hcoil += 1
      if 'Coil:Cooling:DX' in line:
        coilname = get_eplus_token (fp.readline())
        if coilname not in ccoils:
          ccoils[coilname] = {'Sensor': 'Cooling_Coil_{:d}'.format(idx_ccoil)}
          idx_ccoil += 1
      if 'Schedule:Compact' in line:
        schedule = line
        line = fp.readline()
        schedule += line
        name = get_eplus_token (line)
        while True:
          line = fp.readline()
          schedule += line
          if ';' in line:
            break
        schedules[name] = {'Schedule': schedule, 'Used': False, 'Heating': False, 'Alias' : ''}
      if 'ThermostatSetpoint:DualSetpoint' in line:
        name = get_eplus_token (fp.readline())
        heat = get_eplus_token (fp.readline())
        cool = get_eplus_token (fp.readline())
        thermostats[name] = {'Heating': heat, 'Cooling': cool}
        nsetpoints += 1
      if 'ZoneControl:Thermostat' in line:
        name = get_eplus_token (fp.readline())
        zone = get_eplus_token (fp.readline())
        ctrltype = get_eplus_token (fp.readline())
        objtype = get_eplus_token (fp.readline())
        ctrlname = get_eplus_token (fp.readline())
        zonecontrols[zone] = ctrlname
        ncontrols += 1
    line = fp.readline()

  fp.close()

  idx_hsched = 1
  idx_csched = 1
  for zone, ctrlname in zonecontrols.items():
    if ctrlname in thermostats:
      heat = thermostats[ctrlname]['Heating']
      cool = thermostats[ctrlname]['Cooling']
      if not schedules[heat]['Used']:
        schedules[heat]['Used'] = True
        schedules[heat]['Heating'] = True
        schedules[heat]['Alias'] = 'H{:d}'.format(idx_hsched)
        idx_hsched += 1
      if not schedules[cool]['Used']:
        schedules[cool]['Used'] = True
        schedules[cool]['Alias'] = 'C{:d}'.format(idx_csched)
        idx_csched += 1
      zones[zone]['Hsched'] = heat
      zones[zone]['Csched'] = cool
      zones[zone]['Controlled'] = True
    else:
      print ('  ** No Schedule Found for Zone={:s}'.format(zone))
  for name, row in schedules.items():
    if row['Used']:
      nschedused += 1

  nhvacs = len(hvacs)
  nccoils = len(ccoils)
  nhcoils = len(hcoils)
  print ('  === {:d} zones total {:.2f} m3 with {:d} zone controls, {:d} dual setpoints, {:d} schedules, {:d} heating coils, {:d} cooling coils and {:d} HVAC loops'.format(
    nzones, volume, ncontrols, nsetpoints, nschedused, nhcoils, nccoils, nhvacs))
  return zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs

def write_new_ems (target, zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs):
  op = open(target, 'w')
  print ("""! ***EMS PROGRAM***""", file=op)
  nschedused = 0
  for key, row in schedules.items():
    if row['Used']:
      alias = row['Alias'] + '_NOM'
      insensitive = re.compile(re.escape(key), re.IGNORECASE)
      data = insensitive.sub(alias, row['Schedule'])
      print (data, file=op)
      nschedused += 1

  print ("""
  EnergyManagementSystem:ProgramCallingManager,
    Volume_Initializer,   !- Name
    BeginNewEnvironment,  !- Calling Point
    Initialize_Volumes;   !- Program Name
  EnergyManagementSystem:ProgramCallingManager,
    Setpoint_Controller,          !- Name
    BeginTimestepBeforePredictor, !- Calling Point
    Set_Setpoints;                !- Program Name
  EnergyManagementSystem:ProgramCallingManager,
    Demand_Reporter,                      !- Name
    EndOfZoneTimestepAfterZoneReporting,  !- Calling Point
    Report_Demand,
    Report_Occupants,
    Calculate_Temperatures;               !- Program Name
  EnergyManagementSystem:Program,
    Set_Setpoints,      !- Name""", file=op)
  idx = 1
  for key, row in schedules.items():
    if row['Used']:
      alias = row['Alias']
      setp = 'COOL_SETP_DELTA'
      term = ';'
      if 'H' in alias[0]:
        setp = 'HEAT_SETP_DELTA'
      if idx < nschedused:
        term = ','
      print ('    Set {:s} = {:s}_NOM + {:s}*5.0/9.0{:s}'.format (alias, alias, setp, term), file=op)
      idx += 1

  print("""
  EnergyManagementSystem:Program,
    Initialize_Volumes,""", file=op)
  term = ','
  idx = 1
  nzones = len(zones)
  nocczones = 0
  nctrlzones = 0
  total_volume = 0.0
  controlled_volume = 0.0
  for zname, row in zones.items():
    zvol = row['zvol']
    if row['People']:
      nocczones += 1
    if row['Controlled']:
      nctrlzones += 1
      controlled_volume += zvol
    total_volume += zvol
    if idx == nzones:
      term = ';'
    print ('    Set {:s}_V = {:.2f}{:s}'.format (valid_var(zname), zvol, term), file=op)
    idx += 1

  print ("""  
  EnergyManagementSystem:Program,
    Calculate_Temperatures,
    Set TOTAL_COOL_V = 0.0,
    Set TOTAL_HEAT_V = 0.0,
    Set C_SET = 0.0,
    Set H_SET = 0.0,
    Set C_CUR = 0.0,
    Set H_CUR = 0.0,""", file=op)
  print ('    Set Total_V = {:.2f},'.format (total_volume), file=op)
  print ('    Set Controlled_V = {:.2f},'.format (controlled_volume), file=op)

  for zname, row in zones.items():
    if row['Controlled']:
      sname = valid_var(zname)
      Hsens = sname + '_H'
      Hsched = row['Hsched']
      Halias = schedules[Hsched]['Alias']
      Csens = sname + '_C'
      Csched = row['Csched']
      Calias = schedules[Csched]['Alias']
      print ('    IF ({:s} > 0),'.format (Hsens), file=op)
      print ('      Set H_SET = H_SET + {:s} * {:s}_V,'.format (Halias, sname), file=op)
      print ('      Set H_CUR = H_CUR + {:s}_T * {:s}_V,'.format (sname, sname), file=op)
      print ('      Set TOTAL_HEAT_V = TOTAL_HEAT_V + {:s}_V,'.format (sname), file=op)
      print ('    ENDIF,', file=op)
      print ('    IF ({:s} > 0),'.format (Csens), file=op)
      print ('      Set C_SET = C_SET + {:s} * {:s}_V,'.format (Calias, sname), file=op)
      print ('      Set C_CUR = C_CUR + {:s}_T * {:s}_V,'.format (sname, sname), file=op)
      print ('      Set TOTAL_COOL_V = TOTAL_COOL_V + {:s}_V,'.format (sname), file=op)
      print ('    ENDIF,', file=op)

  print ("""! Average temperature over controlled zone air volumes""", file=op)
  print ('    Set T_CUR = 0,', file=op)
  for zname, row in zones.items():
    if row['Controlled']:
      sname = valid_var(zname)
      print ('    Set T_CUR = T_CUR + {:s}_T * {:s}_V,'.format (sname, sname), file=op)
  print ('    Set T_CUR = T_CUR/Controlled_V*9.0/5.0+32.0,', file=op)

  print ("""! Average cooling schedule and setpoint over controlled zone air volumes
    Set Schedule_Cooling_Temperature = 0.0,
    Set T_Cooling = 0,""", file=op)
  for zname, row in zones.items():
    if row['Controlled']:
      sname = valid_var(zname)
      alias = schedules[row['Csched']]['Alias']
      print ('    Set T_Cooling = T_Cooling + {:s} * {:s}_V,'.format (alias, sname), file=op)
      print ('    Set Schedule_Cooling_Temperature = Schedule_Cooling_Temperature + {:s}_NOM * {:s}_V,'.format (alias, sname), file=op)
  print ('    Set T_Cooling = T_Cooling/Controlled_V*9.0/5.0+32.0,', file=op)
  print ('    Set Schedule_Cooling_Temperature = Schedule_Cooling_Temperature/Controlled_V*9.0/5.0+32.0,', file=op)

  print ("""! Average heating schedule and setpoint over controlled zone air volumes
    Set Schedule_Heating_Temperature = 0.0,
    Set T_Heating = 0,""", file=op)
  for zname, row in zones.items():
    if row['Controlled']:
      sname = valid_var(zname)
      alias = schedules[row['Hsched']]['Alias']
      print ('    Set T_Heating = T_Heating + {:s} * {:s}_V,'.format (alias, sname), file=op)
      print ('    Set Schedule_Heating_Temperature = Schedule_Heating_Temperature + {:s}_NOM * {:s}_V,'.format (alias, sname), file=op)
  print ('    Set T_Heating = T_Heating/Controlled_V*9.0/5.0+32.0,', file=op)
  print ('    Set Schedule_Heating_Temperature = Schedule_Heating_Temperature/Controlled_V*9.0/5.0+32.0,', file=op)

  print ("""
    Set Setpoint_Cooling_Temperature = T_Cooling,
    Set Current_Cooling_Temperature = T_CUR,
    Set Setpoint_Heating_Temperature = T_Heating,
    Set Current_Heating_Temperature = T_CUR;
""", file=op)

#  print ("""
#    Set Setpoint_Cooling_Temperature = 0.0,
#    Set Current_Cooling_Temperature = 0.0,
#    Set Setpoint_Heating_Temperature = 0.0,
#    Set Current_Heating_Temperature = 0.0,
#
#    IF (C_SET > 0 && H_SET > 0), ! Scenario 1, both heating and cooling
#      Set Setpoint_Cooling_Temperature = C_SET/TOTAL_COOL_V*9.0/5.0+32.0,
#      Set Setpoint_Heating_Temperature = H_SET/TOTAL_HEAT_V*9.0/5.0+32.0,
#      Set Current_Cooling_Temperature = C_CUR/TOTAL_COOL_V*9.0/5.0+32.0,
#      Set Current_Heating_Temperature = H_CUR/TOTAL_HEAT_V*9.0/5.0+32.0,
#    ELSEIF (C_SET == 0 && H_SET == 0),  ! Scenario 2, no heating or cooling
#      Set Setpoint_Cooling_Temperature = T_Cooling,
#      Set Setpoint_Heating_Temperature = T_Heating,
#      Set Current_Cooling_Temperature = T_Cooling,
#      Set Current_Heating_Temperature = T_Heating,
#    ELSEIF (C_SET > 0 && H_SET == 0), ! Scenario 3, only cooling
#      Set Setpoint_Cooling_Temperature = C_SET/TOTAL_COOL_V*9.0/5.0+32.0,
#      Set Setpoint_Heating_Temperature = T_Heating,
#      Set Current_Cooling_Temperature = C_CUR/TOTAL_COOL_V*9.0/5.0+32.0,
#      Set Current_Heating_Temperature = T_Heating,
#    ELSEIF (C_SET == 0 && H_SET > 0), ! Scenario 4, only heating
#      Set Setpoint_Cooling_Temperature = T_Cooling,
#      Set Setpoint_Heating_Temperature = H_SET/TOTAL_HEAT_V*9.0/5.0+32.0,
#      Set Current_Cooling_Temperature = T_Cooling,
#      Set Current_Heating_Temperature = H_CUR/TOTAL_HEAT_V*9.0/5.0+32.0,
#    ENDIF;
#""", file=op)

  print ("""  
  EnergyManagementSystem:Program,
    Report_Demand,      !- Name
    Set Cooling_Power_State = 0.0,
    Set Heating_Power_State = 0.0,
    Set Flexible_Cooling_Demand = 0.0,
    Set Flexible_Heating_Demand = 0.0,""", file=op)
  for name, row in ccoils.items():
    print ('    Set Flexible_Cooling_Demand = Flexible_Cooling_Demand + {:s},'.format(row['Sensor']), file=op)
  print ('    Set Flexible_Cooling_Demand = Flexible_Cooling_Demand/(60*60*ZoneTimeStep),', file=op)
  print ('    IF Flexible_Cooling_Demand > 1.0,', file=op)
  print ('      Set Cooling_Power_State = 1.0,', file=op)
  print ('    ENDIF,', file=op)
  for name, row in hcoils.items():
    print ('    Set Flexible_Heating_Demand = Flexible_Heating_Demand + {:s},'.format(row['Sensor']), file=op)
  print ('    Set Flexible_Heating_Demand = Flexible_Heating_Demand/(60*60*ZoneTimeStep),', file=op)
  print ('    IF Flexible_Heating_Demand > 1.0,', file=op)
  print ('      Set Heating_Power_State = 1.0,', file=op)
  print ('    ENDIF;', file=op)

  print ("""  
  EnergyManagementSystem:Program,
    Report_Occupants,
    Set Total_Occupants = 0.0,""", file=op)
  term = ','
  idx = 1
  for name, row in zones.items():
    if row['People']:
      if idx == nocczones:
        term = ';'
      print ('    Set Total_Occupants = Total_Occupants + {:s}_O{:s}'.format(valid_var(name), term), file=op)
      idx += 1

  for name, row in schedules.items():
    if row['Used']:
      alias = row['Alias']
      schedule_sensor (alias + '_NOM', op)
      schedule_actuator (alias, name, op)

  global_variable ('Flexible_Cooling_Demand', op)
  global_variable ('Flexible_Heating_Demand', op)
  global_variable ('Setpoint_Cooling_Temperature', op)
  global_variable ('Setpoint_Heating_Temperature', op)
  global_variable ('Schedule_Cooling_Temperature', op)
  global_variable ('Schedule_Heating_Temperature', op)
  global_variable ('Current_Cooling_Temperature', op)
  global_variable ('Current_Heating_Temperature', op)
  global_variable ('Cooling_Power_State', op)
  global_variable ('Heating_Power_State', op)
  global_variable ('H_SET', op)
  global_variable ('C_SET', op)
  global_variable ('H_CUR', op)
  global_variable ('C_CUR', op)
  global_variable ('TOTAL_HEAT_V', op)
  global_variable ('TOTAL_COOL_V', op)
  global_variable ('T_CUR', op)
  global_variable ('Total_Occupants', op)

  output_variable ('Cooling Controlled Load', 'Flexible_Cooling_Demand', op)
  output_variable ('Heating Controlled Load', 'Flexible_Heating_Demand', op)
  output_variable ('Cooling Schedule Temperature', 'Schedule_Cooling_Temperature', op)
  output_variable ('Heating Schedule Temperature', 'Schedule_Heating_Temperature', op)
  output_variable ('Cooling Setpoint Temperature', 'Setpoint_Cooling_Temperature', op)
  output_variable ('Heating Setpoint Temperature', 'Setpoint_Heating_Temperature', op)
  output_variable ('Cooling Current Temperature', 'Current_Cooling_Temperature', op)
  output_variable ('Heating Current Temperature', 'Current_Heating_Temperature', op)
  output_variable ('Cooling Power State', 'Cooling_Power_State', op)
  output_variable ('Heating Power State', 'Heating_Power_State', op)
  output_variable ('Heating Setpoint', 'H_SET', op)
  output_variable ('Cooling Setpoint', 'C_SET', op)
  output_variable ('Heating Current', 'H_CUR', op)
  output_variable ('Cooling Current', 'C_CUR', op)
  output_variable ('Heating Volume', 'TOTAL_HEAT_V', op)
  output_variable ('Cooling Volume', 'TOTAL_COOL_V', op)
  output_variable ('Indoor Air Temperature', 'T_CUR', op)
  output_variable ('Occupant Count', 'Total_Occupants', op)

  for name, row in ccoils.items():
    cooling_coil_sensor (row['Sensor'], name, op)
  for name, row in hcoils.items():
    heating_coil_sensor (row['Sensor'], name, op)

  for zname, row in zones.items():
    if row['People']:
      zone_occupant_sensor (zname, op)
    if row['Controlled']:
      zone_temperature_sensor (zname, op)
      zone_sensible_heating_sensor (zname, op)
      zone_sensible_cooling_sensor (zname, op)
    global_variable (valid_var (zname) + '_V', op)

  print ("""! ***EXTERNAL INTERFACE***
  ExternalInterface,
    FNCS;             !- Name of External Interface
  ExternalInterface:Variable,
    COOL_SETP_DELTA,  !- Name
    0;                !- Initial Value
  ExternalInterface:Variable,
    HEAT_SETP_DELTA,  !- Name
    0;                !- Initial Value
! ***GENERAL REPORTING***
  Output:VariableDictionary,IDF,Unsorted;
! ***REPORT METERS/VARIABLES***
  Output:Variable,EMS,Cooling Controlled Load,timestep;
  Output:Variable,EMS,Heating Controlled Load,timestep;
  Output:Variable,EMS,Cooling Schedule Temperature,timestep;
  Output:Variable,EMS,Heating Schedule Temperature,timestep;
  Output:Variable,EMS,Cooling Setpoint Temperature,timestep;
  Output:Variable,EMS,Heating Setpoint Temperature,timestep;
  Output:Variable,EMS,Cooling Current Temperature,timestep;
  Output:Variable,EMS,Heating Current Temperature,timestep;
  Output:Variable,EMS,Cooling Power State,timestep;
  Output:Variable,EMS,Heating Power State,timestep;
  Output:Variable,EMS,Cooling Volume,timestep;
  Output:Variable,EMS,Heating Volume,timestep;
  Output:Variable,EMS,Occupant Count,timestep;
  Output:Variable,EMS,Indoor Air Temperature,timestep;
  Output:Variable,WHOLE BUILDING,Facility Total Electric Demand Power,timestep;
  Output:Variable,WHOLE BUILDING,Facility Total HVAC Electric Demand Power,timestep;
  Output:Variable,FACILITY,Facility Thermal Comfort ASHRAE 55 Simple Model Summer or Winter Clothes Not Comfortable Time,timestep;
  Output:Variable,*,Site Outdoor Air Drybulb Temperature,timestep; """, file=op)

  op.close()

def make_ems(sourcedir='./output', baseidf='SchoolBase.idf', target='ems.idf'):
  """Creates the EMS for an EnergyPlus building model

  Args:
    target (str): desired output file in PWD, default ems.idf
    baseidf (str): is the original EnergyPlus model file without the EMS
    sourcedir (str): directory of the output from EnergyPlus baseline simulation, default ./output
  """

  print ('*** make_ems from', sourcedir, 'to', target)
  zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs = summarize_idf (sourcedir + '/eplusout.eio', baseidf)
#  print_idf_summary (target, zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs)
  return write_new_ems (target, zones, zonecontrols, thermostats, schedules, hcoils, ccoils, hvacs)

def merge_idf (base, ems, StartTime, EndTime, target, StepsPerHour):
  """Assembles a base EnergyPlus building model with EMS and simulation period

  Args:
    base (str): fully qualified base IDF model
    ems (str): fully qualified EMS model file
    StartTime (str): Date-Time to start simulation, Y-m-d H:M:S
    EndTime (str): Date-Time to end simulation, Y-m-d H:M:S
  """

  time_fmt = '%Y-%m-%d %H:%M:%S'
  dt1 = datetime.strptime (StartTime, time_fmt)
  dt2 = datetime.strptime (EndTime, time_fmt)

  ep_dow_names = ['Monday,   ', 'Tuesday,  ', 'Wednesday,', 'Thursday, ', 'Friday,   ', 'Saturday, ', 'Sunday,   ']
  dow = dt1.weekday()
  begin_month = dt1.month
  begin_day = dt1.day
  end_month = dt2.month
  end_day = dt2.day
  if dt2.hour == 0 and dt2.minute == 0 and dt2.second == 0:
    end_day -= 1

  ip = open (base, 'r', encoding='latin-1')
  op = open (target, 'w', encoding='latin-1')
  print ('filtering', base, 'plus', ems, 'to', target)
  for ln in ip:
    line = ln.rstrip('\n')
    if '!- Begin Month' in line:
      print ('    %s                      !- Begin Month' % helpers.idf_int(begin_month), file=op)
    elif '!- Begin Day of Month' in line:
      print ('    %s                      !- Begin Day of Month' % helpers.idf_int(begin_day), file=op)
    elif '!- End Month' in line:
      print ('    %s                      !- End Month' % helpers.idf_int(end_month), file=op)
    elif '!- End Day of Month' in line:
      print ('    %s                      !- End Day of Month' % helpers.idf_int(end_day), file=op)
    elif '!- Day of Week for Start Day' in line:
      print ('    %s               !- Day of Week for Start Day' % ep_dow_names[dow], file=op)
    elif 'Timestep,' in line:
      print ('  Timestep,%s;' % str(StepsPerHour), file=op)
    else:
      print (line, file=op)
  ip.close()

  ip = open (ems, 'r', encoding='latin-1')
  for ln in ip:
    line = ln.rstrip('\n')
    print (line, file=op)
  op.close()

